# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import math
import time
from pyspades.vxl cimport VXLData, MapData
from pyspades.common cimport Vertex3, create_proxy_vector
from libc.math cimport sqrt, sin, cos, acos, fabs
from pyspades.constants import TORSO, HEAD, ARMS, LEGS, MELEE

cdef extern from "common_c.h":
    struct LongVector:
        int x, y, z
    struct Vector:
        float x, y, z

cdef extern from "world_c.cpp":
    enum:
        CUBE_ARRAY_LENGTH
    int c_validate_hit "validate_hit" (
        float shooter_x, float shooter_y, float shooter_z,
        float orientation_x, float orientation_y, float orientation_z,
        float victim_x, float victim_y, float victim_z, float tolerance)
    int c_can_see "can_see" (MapData * map, float x0, float y0, float z0,
        float x1, float y1, float z1)
    int c_cast_ray "cast_ray" (MapData * map, float x0, float y0, float z0,
        float x1, float y1, float z1, float length, long* x, long* y, long* z)
    size_t cube_line_c "cube_line"(int, int, int, int, int, int, LongVector *)
    void set_globals(MapData * map, float total_time, float dt)
    struct PlayerType:
        Vector p, e, v, s, h, f
        int mf, mb, ml, mr
        int jump, crouch, sneak
        int airborne, wade, alive, sprint
        int primary_fire, secondary_fire, weapon

    struct GrenadeType:
        Vector p, v
    PlayerType * create_player()
    void destroy_player(PlayerType * player)
    void destroy_grenade(GrenadeType * player)
    void update_timer(float value, float dt)
    void reorient_player(PlayerType * p, Vector * vector)
    int move_player(PlayerType * p)
    int try_uncrouch(PlayerType * p)
    GrenadeType * create_grenade(Vector * p, Vector * v)
    int move_grenade(GrenadeType * grenade)

from libc.math cimport sqrt

cdef inline bint can_see(VXLData map, float x1, float y1, float z1,
    float x2, float y2, float z2):
    return c_can_see(map.map, x1, y1, z1, x2, y2, z2)

cdef inline bint cast_ray(VXLData map, float x1, float y1, float z1,
    float x2, float y2, float z2, float length, long* x, long* y, long* z):
    return c_cast_ray(map.map, x1, y1, z1, x2, y2, z2, length, x, y, z)

cdef class Object
cdef class World
cdef class Grenade
cdef class Character

cdef class Object:
    """an object in present in the World"""
    cdef public:
        object name
        World world

    def __init__(self, world, *arg, **kw):
        self.world = world
        self.initialize(*arg, **kw)
        if self.name is None:
            self.name = 'object'

    def initialize(self, *arg, **kw):
        """hook called on Object creation

        Arguments passed to ``__init__`` will be passed here too.
        """
        pass

    cdef int update(self, double dt) except -1:
        '''update this object, giving it a "Tick"'''
        return 0

    def delete(self):
        """remove this object from the World"""
        self.world.delete_object(self)

cdef class Character(Object):
    """Represents the position, orientation and velocity of the player object in
    the world"""
    cdef:
        PlayerType * player
    cdef public:
        Vertex3 position, orientation, velocity
        object fall_callback

    def initialize(self, Vertex3 position, Vertex3 orientation,
                   fall_callback = None):
        self.name = 'character'
        self.player = create_player()
        self.fall_callback = fall_callback
        self.position = create_proxy_vector(&self.player.p)
        self.orientation = create_proxy_vector(&self.player.f)
        self.velocity = create_proxy_vector(&self.player.v)
        if position is not None:
            self.set_position(*position.get())
        if orientation is not None:
            self.orientation.set_vector(orientation)

    def set_crouch(self, bint value):
        """set if the player is crouching"""
        if value == self.player.crouch:
            return
        if value:
            self.player.p.z += 0.9
        else:
            self.player.p.z -= 0.9
        self.player.crouch = value

    def set_animation(self, jump, crouch, sneak, sprint):
        """set all of the player's movement statuses: jump, crouch, sneak and
        sprint"""
        self.player.jump = jump
        self.set_crouch(crouch)
        self.player.sneak = sneak
        self.player.sprint = sprint

    def set_weapon(self, is_primary):
        """set the primary weapon of the player"""
        self.player.weapon = is_primary

    def set_walk(self, up, down, left, right):
        """set the current status of the movement buttons"""
        self.player.mf = up
        self.player.mb = down
        self.player.ml = left
        self.player.mr = right

    def set_position(self, x, y, z, reset = False):
        """set the current position of the player. If ``reset=True`` is passed,
        reset velocity, keys, mouse buttons and movement status as well"""
        self.position.set(x, y, z)
        self.player.p.x = self.player.e.x = x
        self.player.p.y = self.player.e.y = y
        self.player.p.z = self.player.e.z = z
        if reset:
            self.velocity.set(0.0, 0.0, 0.0)
            self.primary_fire = self.secondary_fire = False
            self.jump = self.crouch = False
            self.up = self.down = self.left = self.right = False

    def set_orientation(self, x, y, z):
        """set the current orientation of the Player"""
        cdef Vertex3 v = Vertex3(x, y, z)
        reorient_player(self.player, v.value)

    cpdef int can_see(self, float x, float y, float z):
        """return if the player can see a given coordinate. This only considers
        the map voxels, not any other objects"""
        cdef Vertex3 position = self.position
        return can_see(self.world.map, position.x, position.y, position.z,
            x, y, z)

    cpdef cast_ray(self, length = 32.0):
        """cast a ray ``length`` number of blocks in the direction the player is
        facing, If a voxel is hit, return it's coordinates, otherwise `None`"""
        cdef Vertex3 position = self.position
        cdef Vertex3 direction = self.orientation.copy().normal()
        cdef long x, y, z
        if cast_ray(self.world.map, position.x, position.y, position.z,
            direction.x, direction.y, direction.z, length, &x, &y, &z):
            return x, y, z
        return None

    def validate_hit(self, Character other, part, float tolerance):
        """check if a given hit is within a given tolerance of hitting another
        player. This is primarily used to prevent players from shooting at
        things they aren't facing at"""
        cdef Vertex3 position1 = self.position
        cdef Vertex3 orientation = self.orientation
        cdef Vertex3 position2 = other.position
        cdef float x, y, z
        x = position2.x
        y = position2.y
        z = position2.z
        if part in (TORSO, ARMS):
            z += 0.9
        elif part == HEAD:
            pass
        elif part == LEGS:
            z += 1.8
        elif part == MELEE:
            z += 0.9
        else:
            return False
        if not c_validate_hit(position1.x, position1.y, position1.z,
                              orientation.x, orientation.y, orientation.z,
                              x, y, z, tolerance):
            return False
        return True

    def set_dead(self, value):
        """set the player's alive status. Also resets mouse buttons, movement
        stats and keys"""
        self.player.alive = not value
        self.player.mf = False
        self.player.mb = False
        self.player.ml = False
        self.player.mr = False
        self.player.crouch = False
        self.player.sneak = False
        self.player.primary_fire = False
        self.player.secondary_fire = False
        self.player.sprint = False

    cdef int update(self, double dt) except -1:
        cdef long ret = move_player(self.player)
        if ret > 0:
            self.fall_callback(ret)
        return 0

    # properties
    property up:
        def __get__(self):
            return self.player.mf
        def __set__(self, value):
            self.player.mf = value

    property down:
        def __get__(self):
            return self.player.mb
        def __set__(self, value):
            self.player.mb = value

    property left:
        def __get__(self):
            return self.player.ml
        def __set__(self, value):
            self.player.ml = value

    property right:
        def __get__(self):
            return self.player.mr
        def __set__(self, value):
            self.player.mr = value

    property dead:
        def __get__(self):
            return not self.player.alive
        def __set__(self, bint value):
            self.set_dead(value)

    property jump:
        def __get__(self):
            return self.player.jump
        def __set__(self, value):
            self.player.jump = value

    property airborne:
        def __get__(self):
            return self.player.airborne

    property crouch:
        def __get__(self):
            return self.player.crouch
        def __set__(self, value):
            self.player.crouch = value

    property sneak:
        def __get__(self):
            return self.player.sneak
        def __set__(self, value):
            self.player.sneak = value

    property wade:
        def __get__(self):
            return self.player.wade

    property sprint:
        def __get__(self):
            return self.player.sprint
        def __set__(self, value):
            self.player.sprint = value

    property primary_fire:
        def __get__(self):
            return self.player.primary_fire
        def __set__(self, value):
            self.player.primary_fire = value

    property secondary_fire:
        def __get__(self):
            return self.player.secondary_fire
        def __set__(self, value):
            self.player.secondary_fire = value

    def __repr__(self):
        return f"Character(pos={self.position}, orientation={self.orientation})"

cdef class Grenade(Object):
    cdef public:
        Vertex3 position, velocity
        float fuse
        object callback
        object team
    cdef GrenadeType * grenade

    def initialize(self, double fuse, Vertex3 position, Vertex3 orientation,
                   Vertex3 velocity, callback = None):
        self.name = 'grenade'
        self.grenade = create_grenade(position.value, velocity.value)
        self.position = create_proxy_vector(&self.grenade.p)
        self.velocity = create_proxy_vector(&self.grenade.v)
        if orientation is not None:
            self.velocity += orientation
        self.fuse = fuse
        self.callback = callback

    cdef int hit_test(self, Vertex3 position):
        """check if the grenade can hit something in a given position. This is
        used to check if the player should be damaged"""
        cdef Vector * nade = self.position.value
        return can_see(self.world.map, position.x, position.y, position.z,
                       nade.x, nade.y, nade.z)

    cpdef get_next_collision(self, double dt):
        """calculate the position of the grenade ahead of time.

        Returns:
            eta, x, y, z: the ETA and the location of the next collision
        """
        if self.velocity.is_zero():
            return None
        cdef double eta = 0.0
        cdef double x, y, z
        cdef Vertex3 old_position = self.position.copy()
        cdef Vertex3 old_velocity = self.velocity.copy()
        while move_grenade(self.grenade) == 0:
            eta += dt
            if eta > 5.0:
                break
        x, y, z = self.position.x, self.position.y, self.position.z
        self.position.set_vector(old_position)
        self.velocity.set_vector(old_velocity)
        return eta, x, y, z

    cpdef double get_damage(self, Vertex3 player_position):
        """Calculate the damage given to a player standing at
        ``player_position``. Also performs a check to see if the player is
        behind cover."""
        cdef Vector * position = self.position.value
        cdef double diff_x, diff_y, diff_z
        diff_x = player_position.x - position.x
        diff_y = player_position.y - position.y
        diff_z = player_position.z - position.z
        cdef double value
        if (fabs(diff_x) < 16 and
            fabs(diff_y) < 16 and
            fabs(diff_z) < 16 and
            self.hit_test(player_position)):
            value = diff_x**2 + diff_y**2 + diff_z**2
            if value == 0.0:
                return 100.0
            return 4096.0 / value
        return 0

    cdef int update(self, double dt) except -1:
        self.fuse -= dt
        if self.fuse <= 0:
            if self.callback is not None:
                self.callback(self)
            self.delete()
            return 0
        move_grenade(self.grenade)

    def __dealloc__(self):
        destroy_grenade(self.grenade)

    def __repr__(self):
        rep = "Grenade(fuse={:.3f}, position={}, (...), velocity={})"
        return rep.format(self.fuse, self.position, self.velocity)

cdef class World(object):
    """controls the map of the World and the Objects inside of it"""
    cdef public:
        VXLData map
        list objects
        float time

    def __init__(self):
        self.objects = []
        self.time = 0

    def update(self, double dt):
        if self.map is None:
            return
        self.time += dt
        set_globals(self.map.map, self.time, dt)
        cdef Object instance
        for instance in self.objects[:]:
            instance.update(dt)

    cpdef delete_object(self, Object item):
        self.objects.remove(item)

    def create_object(self, klass, *arg, **kw):
        new_object = klass(self, *arg, **kw)
        self.objects.append(new_object)
        return new_object

# utility functions

cpdef cube_line(x1, y1, z1, x2, y2, z2):
    """create a cube line from one point to another with the same algorithm as
    the client uses"""
    cdef LongVector array[CUBE_ARRAY_LENGTH]
    cdef size_t size = cube_line_c(x1, y1, z1, x2, y2, z2, array)
    cdef size_t i
    cdef list points = []
    for i in range(size):
        points.append((array[i].x, array[i].y, array[i].z))
    return points



"""
Badmin is an bot admin.

It automates common admin tasks such as:

* Banning griefers
* Banning aimbotters
* Banning racists

Since it is automatic it won't always get it right.

Commands
^^^^^^^^

* ``/badmin`` shows which badmin options are enabled/disabled *admin only*
* ``/investigate <player>`` shows a player's grief score, K/D ratio and hit accuracy *admin only*

.. note::
  It depends on the blockinfo.py (for grief detection), ratio.py (for k/d ratio), aimbot2.py (hit accuracy) scripts.

.. codeauthor:: ?
"""

from twisted.internet import reactor
from pyspades.common import prettify_timespan
from pyspades.constants import *
from pyspades.collision import distance_3d_vector
from piqueserver.commands import command, admin, get_player
import re

# TODO: convert settings to config options
BADMIN_VERSION = 9
# Settings for auto-aimbot
SCORE_AIMBOT_ENABLED = True
# any votekicks under uncertain will be cancelled
SCORE_AIMBOT_UNCERTAIN = 2
SCORE_AIMBOT_WARN = 5
SCORE_AIMBOT_KICK = 15
SCORE_AIMBOT_BAN = 25

# Settings for auto-griefcheck
SCORE_GRIEF_ENABLED = True
# any votekicks under uncertain will be cancelled
SCORE_GRIEF_UNCERTAIN = 2
SCORE_GRIEF_WARN = 4
SCORE_GRIEF_KICK = 8
SCORE_GRIEF_BAN = 12

# Settings for blank reason votekicks
# turns on setting preventing blank votekicks
BLANK_VOTEKICK_ENABLED = True

# Settings for language filter
LANGUAGE_FILTER_ENABLED = True


slur_pattern = re.compile(".*nigger.*", re.IGNORECASE)
grief_pattern = re.compile(
    ".*(gr.*f.*(ing|er)|grief|destroy).*", re.IGNORECASE)
aimbot_pattern = re.compile(".*(aim|bot|ha(ck|x)|cheat).*", re.IGNORECASE)


def slur_match(player, msg):
    return (not slur_pattern.match(msg) is None)


def grief_match(player, msg):
    return (not grief_pattern.match(msg) is None)


def aimbot_match(player, msg):
    return (not aimbot_pattern.match(msg) is None)


@command(admin_only=True)
def badmin(connection, var=None):
    if var is None:
        return (
            "@Badmin (r%s): Language Filter(LF) [%s], Blank Votekick Blocker(BV) "
            "[%s], Grief Votekick Protection(GV) [%s], Aimbot Votekick Protection(AV) [%s]" %
            (BADMIN_VERSION,
             LANGUAGE_FILTER_ENABLED,
             BLANK_VOTEKICK_ENABLED,
             SCORE_GRIEF_ENABLED,
             SCORE_AIMBOT_ENABLED))


@command(admin_only=True)
def investigate(connection, player):
    player = get_player(connection.protocol, player)
    score = score_grief(connection, player)
    kdr = round(player.ratio_kills / float(max(1, player.ratio_deaths)))
    percent = round(check_percent(player))
    message = "Results for %s: Grief Score - %s / KDR - %s / Hit Acc. - %s" % (
        player.name, score, kdr, percent)
    return message


def score_grief(connection, player, time=None):  # 302 = blue (0), #303 = green (1)
    print("start score grief")
    color = connection not in connection.protocol.players.values() and connection.colors
    minutes = float(time or 2)
    if minutes < 0.0:
        raise ValueError("Minutes cannot be < 0")
    time = reactor.seconds() - minutes * 60.0
    blocks_removed = player.blocks_removed or []
    blocks = [b[1] for b in blocks_removed if b[0] >= time]
    player_name = player.name
    team_id = player.team.id  # 0=blue, 1=green
    print("name/team set")
    gscore = 0  # griefscore
    map_blocks = 0
    team_blocks = 0
    enemy_blocks = 0
    team_harmed = 0
    enemy_harmed = 0
    print("init values set")
    if len(blocks):
        print("len blocks = true, blocks found")
        total_blocks = len(blocks)
        info = blocks
        for info in blocks:
            if info:
                name, team = info
                if name != player_name and team == team_id:
                    team_blocks += 1
                elif team != team_id:
                    enemy_blocks += 1
            else:
                map_blocks += 1
        print("second for done")
        infos = set(blocks)
        infos.discard(None)
        for name, team in infos:
            if name != player_name and team == team_id:
                team_harmed += 1
            elif team != team_id:
                enemy_harmed += 1
        print("third for done")
    else:
        print("len blocks = false, no blocks found")
        total_blocks = 0

    # heuristic checks start here
    # if they didn't break any blocks at all, they probably aren't griefing.
    if total_blocks == 0:
        print("no blocks, ending")
        return 0
    # checks on team blocks destroyed
    if team_blocks > 0 and team_blocks <= 5:
        gscore += 1
    elif team_blocks > 5 and team_blocks <= 10:
        gscore += 2
    elif team_blocks > 10 and team_blocks <= 25:
        gscore += 4
    elif team_blocks > 25 and team_blocks <= 50:
        gscore += 6
    elif team_blocks > 50:
        gscore += 10
    print("team blocks set")
    # team / total ratio checks
    if total_blocks != 0:
        ttr = (float(team_blocks) / float(total_blocks)) * 100
    if ttr > 5 and ttr <= 20:
        gscore += 1
    elif ttr > 20 and ttr <= 50:
        gscore += 2
    elif ttr > 50 and ttr <= 80:
        gscore += 3
    elif ttr > 80:
        gscore += 4
    print("ttr set")
    # teammates harmed check
    if team_harmed == 1:
        gscore += 1
    elif team_harmed > 2 and team_harmed <= 4:
        gscore += 3
    elif team_harmed > 4:
        gscore += 6
    print("team harmed set")
    print(
        "mb: %s, tb: %s, eb: %s, Tb: %s, th: %s, ttr: %s, eh: %s, gs: %s" %
        (map_blocks,
         team_blocks,
         enemy_blocks,
         total_blocks,
         team_harmed,
         ttr,
         enemy_harmed,
         gscore))
    return gscore


def check_percent(self):
    # TODO: uncomment this when you find out what/where semi_hits and
    #       semi_count should be defined. Possibly in aimbot2.py, but currently these
    #       attributes don't exist anywhere
    # if self.weapon == RIFLE_WEAPON:
    #     if self.semi_hits == 0 or self.semi_count == 0:
    #         return 0
    #     else:
    #         return (float(self.semi_hits) / float(self.semi_count)) * 100
    # elif self.weapon == SMG_WEAPON:
    if self.weapon == SMG_WEAPON:
        if self.smg_hits == 0 or self.smg_count == 0:
            return 0
        else:
            return (float(self.smg_hits) / float(self.smg_count)) * 100
    elif self.weapon == SHOTGUN_WEAPON:
        if self.shotgun_hits == 0 or self.shotgun_count == 0:
            return 0
        else:
            return float(self.shotgun_hits) / float(self.shotgun_count)


def apply_script(protocol, connection, config):
    def send_slur_nick(connection):
        badmin_punish(connection, 'kick', 'Being a racist')

    def badmin_punish(
            connection,
            punishment='warn',
            reason="Being a meany face"):
        connection.protocol.irc_say(
            "* @Badmin: %s is being punished. Type: %s (Reason: %s)" %
            (connection.name, punishment, reason))
        if punishment == "ban":
            connection.ban('@Badmin: ' + reason,
                           connection.protocol.votekick_ban_duration)
        elif punishment == "kick":
            connection.kick('@Badmin: ' + reason)
        elif punishment == "warn":
            connection.protocol.send_chat(
                " @Badmin: Hey %s, %s" % (connection.name, reason))

    class BadminConnection(connection):

        def on_chat(self, value, global_message):
            if slur_match(self, value) and LANGUAGE_FILTER_ENABLED:
                reactor.callLater(1.0, send_slur_nick, self)
            return connection.on_chat(self, value, global_message)

    class BadminProtocol(protocol):

        def on_votekick_start(self, connection, player, reason=None):
            if reason is None and BLANK_VOTEKICK_ENABLED:
                connection.protocol.irc_say(
                    "* @Badmin: %s is attempting a blank votekick (against %s)" %
                    (connection.name, player.name))
                return "@Badmin: You must input a reason for the votekick (/votekick name reason)"
            # print "before aimbot check"
            # print player.ratio_kills/float(max(1,player.ratio_deaths))
            if aimbot_match(self, reason) and SCORE_AIMBOT_ENABLED:
                # print "made aimbot check"
                score = round(player.ratio_kills /
                              float(max(1, player.ratio_deaths)))
                percent = round(check_percent(player))
                # print "score: %s, acc: %s" % (score, percent)
                if score >= SCORE_AIMBOT_BAN:
                    badmin_punish(
                        player, "ban", "Suspected Aimbotting (Kicker: %s, KDR: %s, Hit Acc: %s)" %
                        (connection.name, score, percent))
                    return
                if score >= SCORE_AIMBOT_KICK:
                    badmin_punish(
                        player, "kick", "Suspected Aimbotting (Kicker: %s, KDR: %s, Hit Acc: %s)" %
                        (connection.name, score, percent))
                    return
                if score >= SCORE_AIMBOT_WARN:
                    badmin_punish(
                        player, "warn", "People think you're aimbotting! (KDR: %s, Hit Acc: %s)" %
                        (score, percent))
                    return protocol.on_votekick_start(
                        self, connection, player, reason)
                if score >= SCORE_AIMBOT_UNCERTAIN:
                    connection.protocol.irc_say(
                        "* @Badmin: Aimbot vote: (KDR: %s, Hit Acc: %s)" %
                        (score, percent))
                    return protocol.on_votekick_start(
                        self, connection, player, reason)
                if score < SCORE_AIMBOT_UNCERTAIN:
                    connection.protocol.irc_say(
                        "* @Badmin: I've cancelled an aimbot votekick! Kicker: %s, Kickee: %s, KDR: %s, Hit Acc: %s" %
                        (connection.name, player.name, score, percent))
                    return "@Badmin: This player is not aimbotting."
                    # print "went too far (aimbot)"
            if grief_match(self, reason) and SCORE_GRIEF_ENABLED:
                # print "made grief check"
                score = score_grief(connection, player)
                if score >= SCORE_GRIEF_BAN:
                    badmin_punish(
                        player, "ban", "Griefing (Kicker: %s, GS: %s)" %
                        (connection.name, score))
                    return
                if score >= SCORE_GRIEF_KICK:
                    badmin_punish(
                        player, "kick", "Griefing (Kicker: %s, GS: %s)" %
                        (connection.name, score))
                    return
                if score >= SCORE_GRIEF_WARN:
                    badmin_punish(player, "warn",
                                  "Stop Griefing! (GS: %s)" % score)
                    return protocol.on_votekick_start(
                        self, connection, player, reason)
                if score >= SCORE_GRIEF_UNCERTAIN:
                    connection.protocol.irc_say(
                        "* @Badmin: Grief Score: %s" % score)
                    return protocol.on_votekick_start(
                        self, connection, player, reason)
                if score < SCORE_GRIEF_UNCERTAIN:
                    connection.protocol.irc_say(
                        "* @Badmin: I've cancelled a griefing votekick! Kicker: %s, Kickee: %s, Score: %s" %
                        (connection.name, player.name, score))
                    return "@Badmin: This player has not been griefing."
            return protocol.on_votekick_start(self, connection, player, reason)

    return BadminProtocol, BadminConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

"""
Implementation of the 0,75 master server protocol
"""

from pyspades.loaders import Loader
from pyspades.protocol import BaseConnection
from pyspades.constants import MASTER_VERSION

from twisted.internet.defer import Deferred

PORT = 32886

MAX_SERVER_NAME_SIZE = 31
MAX_MAP_NAME_SIZE = 20
MAX_GAME_MODE_SIZE = 7

HOST = 'master.buildandshoot.com'


class AddServer(Loader):
    """The AddServer packet sent to the master server"""
    __slots__ = ['count', 'max_players', 'name', 'port', 'game_mode', 'map']

    id = 4

    def read(self, reader):
        if reader.dataLeft() == 1:
            self.count = reader.readByte(True)
        else:
            self.max_players = reader.readByte(True)
            self.port = reader.readShort(True, False)
            self.name = reader.readString()
            self.game_mode = reader.readString()
            self.map = reader.readString()

    def write(self, writer):
        if self.count is None:
            writer.writeByte(self.max_players)
            writer.writeShort(self.port, True, False)
            writer.writeString(self.name)
            writer.writeString(self.game_mode)
            writer.writeString(self.map)
        else:
            writer.writeByte(self.count, True)


add_server = AddServer()


class MasterConnection(BaseConnection):
    disconnect_callback = None
    connected = False

    def on_connect(self):
        self.connected = True
        self.send_server()

        if self.defer is not None:
            self.defer.callback(self)
            self.defer = None

    def set_count(self, value):
        add_server.count = value
        self.send_contained(add_server)

    def send_server(self):
        protocol = self.server_protocol
        add_server.count = None
        add_server.name = protocol.name[:MAX_SERVER_NAME_SIZE].encode()
        add_server.game_mode = protocol.get_mode_name()[:MAX_GAME_MODE_SIZE].encode()
        add_server.map = protocol.map_info.short_name[:MAX_MAP_NAME_SIZE].encode()
        add_server.port = protocol.host.address.port
        add_server.max_players = protocol.max_players
        self.send_contained(add_server)

    def on_disconnect(self):
        if self.defer is not None:
            self.defer.errback(self)
            self.defer = None
        callback = self.disconnect_callback
        if callback is not None:
            callback()
        self.disconnect_callback = None


def get_master_connection(protocol):
    defer = Deferred()
    connection = protocol.connect(MasterConnection, HOST, PORT, MASTER_VERSION)
    connection.server_protocol = protocol
    connection.defer = defer
    return defer



"""
push.py last modified 2019-03-06
Contributors: danhezee, StackOverflow, izzy, Danke, noway421, IAmYourFriend

The concept:
    Each team spawns at a set location with the enemy intel. They must "push"
    the intel towards their control point, which is also at a set location.
    The only way to arrive there is by building bridges over the deadly water.
    Further introduction to the game mode: https://youtu.be/DdisPY6vDD0

How to setup new maps:
    Spawn and CP locations must be configured via extensions in the map's
    map_name.txt metadata. Example:

>>> extensions = {
...     'push': True,
...     'push_spawn_range' : 5,
...     'push_blue_spawn' : (91, 276, 59),
...     'push_blue_cp' : (91, 276, 59),
...     'push_green_spawn' : (78, 86, 59),
...     'push_green_cp' : (78, 86, 59),
...     'water_damage' : 100
... }

Additional (but optional) extensions, to mark each teams build area and prevent
the enemy from building there (and thereby helping the enemy). The build area
is defined by x and y of upper left corner, followed by x and y of bottom right
corner on the map. Example:

    'push_blue_build_area' : (64, 100, 243, 500),
    'push_green_build_area' : (268, 100, 447, 500),

Options
^^^^^^^

.. code-block:: guess
    [push]
    # Disallow removal of map blocks. This allows a larger variety of maps that
    # rely on more fragile structures. It also prevents griefing (like removing
    # the map blocks before and after your teams bridge). Using server setting
    # 'user_blocks_only' instead doesn't work reliable.
    protect_map_blocks = true

    # Allow the usage of /r to quickly respawn. As players can't refill blocks at
    # their base, they would have to suicide otherwise. This is illogical, messes
    # up their kill-death ratio and gives them an undeserved punishing respawn time.
    allow_respawn_command = true

    # How long to wait to allow the command /r again
    respawn_cmd_delay = "15sec"

    # Players have to wait this amount after spawning before they can pick
    # the intel up. This is to reduce the instant/careless intel pickups.
    intel_pickup_delay = "3sec"

    # How long can you remove your own last blocks
    block_removal_delay = "15sec"

    # Reset intel after it was dropped somewhere
    reset_intel_after_drop = "3min"

    # No building near cp within this block range (can be overwritten using
    # map extension parameter "push_cp_protect_range")
    default_cp_protect_range = 8

    # Disable grenade damage within enemy spawn.
    disable_grenades_at_spawn = false
"""

from pyspades.constants import *
from pyspades.common import make_color
from pyspades.contained import SetColor
from piqueserver.commands import command, admin, get_team
from piqueserver.config import config, cast_duration
from twisted.internet.task import LoopingCall
from random import randint
import colorsys
import time

PUSH_CONFIG = config.section("push")
PROTECT_MAP_BLOCKS = PUSH_CONFIG.option("protect_map_blocks", default=True, cast=bool)
ALLOW_RESPAWN_COMMAND = PUSH_CONFIG.option("allow_respawn_command", default=True, cast=bool)
RESPAWN_CMD_DELAY = PUSH_CONFIG.option("respawn_cmd_delay", default="15sec", cast=cast_duration)
INTEL_PICKUP_DELAY = PUSH_CONFIG.option("intel_pickup_delay", default="3sec", cast=cast_duration)
BLOCK_REMOVAL_DELAY = PUSH_CONFIG.option("block_removal_delay", default="15sec",
                                         cast=cast_duration)
RESET_INTEL_AFTER_DROP = PUSH_CONFIG.option("reset_intel_after_drop", default="3min",
                                            cast=cast_duration)
DEFAULT_CP_PROTECT_RANGE = PUSH_CONFIG.option("default_cp_protect_range", default=8, cast=int)
DISABLE_GRENADES_AT_SPAWN = PUSH_CONFIG.option("disable_grenades_at_spawn", default=False,
                                               cast=bool)

CANT_DESTROY = "You can't destroy your team's blocks!"
NO_BLOCKS = "Out of blocks! Refill at base or type /r"
BUILDING_AT_CP = "You can't build near your base!"
BUILDING_AT_ENEMY_AREA = "Don't build for your enemy!"


def get_now_in_secs():
    return int(time.time())


def byte_rgb_to_hls(rgb):
    hls = colorsys.rgb_to_hls(*tuple(c / 255.0 for c in rgb))
    return tuple(int(round(c * 255)) for c in hls)


def byte_hls_to_rgb(hls):
    rgb = colorsys.hls_to_rgb(*tuple(c / 255.0 for c in hls))
    return tuple(int(round(c * 255)) for c in rgb)


def compare_hs(block_hls, team_hls):
    # if hue and saturation match
    return block_hls[0] == team_hls[0] and block_hls[2] == team_hls[2]


def byte_middle_range(byte):
    half = 50 / 2.0  # half of (byte/5.1)
    min = byte - half
    max = byte + half
    if min < 0:
        min = 0
        max = half
    elif max > 255:
        min = 255 - half
        max = 255
    return int(round(min)), int(round(max))


def create_area(x, y, block_range):
    return (x - block_range, y - block_range, x + block_range, y + block_range)


def is_in_area(x, y, top_x, top_y, bottom_x, bottom_y):
    return top_x <= x < bottom_x and top_y <= y < bottom_y


def has_flag(connection):
    flag = connection.team.other.flag
    if flag.player is not None:
        if flag.player is connection:
            return True
    return False


def reset_intel_position(protocol, team):
    # Flag should always spawn on z-top to prevent griefers burying it under blocks
    pos = (team.other.spawn[0],
           team.other.spawn[1],
           protocol.map.get_z(team.other.spawn[0], team.other.spawn[1], 1))
    team.flag.set(*pos)  # If spawn not set, it would throw error.
    team.flag.update()
    protocol.send_chat("The %s intel has been reset." % team.name)


@command(admin_only=True)
def resetintel(connection, value):
    team = get_team(connection, value)
    reset_intel_position(connection.protocol, team)


@command('r')
def respawn(connection):
    if not ALLOW_RESPAWN_COMMAND.get():
        return
    if connection.world_object is not None and not connection.world_object.dead:
        if (connection.last_spawn_time is None or connection.last_spawn_time +
                RESPAWN_CMD_DELAY.get() <= get_now_in_secs()):
            if has_flag(connection):
                connection.drop_flag()
            connection.spawn()
        else:
            connection.send_chat(
                "Please wait %s seconds before using this command again."
                % (connection.last_spawn_time + RESPAWN_CMD_DELAY.get() - get_now_in_secs()))


def get_entity_location(connection, entity_id):
    if entity_id == BLUE_BASE:
        return connection.protocol.blue_team.cp
    elif entity_id == GREEN_BASE:
        return connection.protocol.green_team.cp

    elif entity_id == BLUE_FLAG:
        return (connection.protocol.green_team.spawn[0],
                connection.protocol.green_team.spawn[1],
                1)
    elif entity_id == GREEN_FLAG:
        return (connection.protocol.blue_team.spawn[0],
                connection.protocol.blue_team.spawn[1],
                1)


def get_spawn_location(connection):
    # distance from spawn center to randomly spawn in
    spawn_range = connection.protocol.spawn_range
    xb = connection.team.spawn[0]
    yb = connection.team.spawn[1]
    xb += randint(-spawn_range, spawn_range)
    yb += randint(-spawn_range, spawn_range)
    return (xb, yb, connection.protocol.map.get_z(xb, yb))


def apply_script(protocol, connection, config):
    class PushConnection(connection):
        last_spawn_time = None
        # list entry format: ((x, y, z), timestamp when block was placed)
        last_blocks = None

        def is_in_invalid_area(self, x, y, check_area, error_message):
            if is_in_area(x, y, *check_area):
                self.send_chat(error_message)
                return True
            else:
                return False

        def invalid_build_position(self, x, y, z):
            # prevent teams from building near their cp
            if self.is_in_invalid_area(x, y, create_area(self.team.cp[0], self.team.cp[1],
                                       self.protocol.cp_protect_range), BUILDING_AT_CP):
                return True
            # prevent teams from building in enemy build area
            if self.team.build_area is not None and self.is_in_invalid_area(
                        x, y, self.team.other.build_area, BUILDING_AT_ENEMY_AREA):
                return True
            return False

        def random_color(self):
            (h, l, s) = self.team.hls
            l = randint(self.team.light_range[0], self.team.light_range[1])
            color = byte_hls_to_rgb((h, l, s))

            self.color = color
            set_color = SetColor()
            set_color.player_id = self.player_id
            set_color.value = make_color(*color)
            self.send_contained(set_color)
            self.protocol.broadcast_contained(set_color, save=True)

        def on_line_build_attempt(self, points):
            can_build = connection.on_line_build_attempt(self, points)
            if can_build is False:
                return False

            if ALLOW_RESPAWN_COMMAND.get() and self.blocks == len(points):
                self.send_chat(NO_BLOCKS)

            for point in points:
                if self.invalid_build_position(*point):
                    return False

            if self.last_blocks is None:
                self.last_blocks = []
            if BLOCK_REMOVAL_DELAY.get() > 0:
                for point in points:
                    x, y, z = point[0], point[1], point[2]
                    if not self.protocol.map.get_solid(x, y, z):
                        self.last_blocks.append(((x, y, z), get_now_in_secs()))

            self.random_color()
            return can_build

        def on_block_build_attempt(self, x, y, z):
            can_build = connection.on_block_build_attempt(self, x, y, z)
            if can_build is False:
                return False

            if ALLOW_RESPAWN_COMMAND.get() and self.blocks == 0:
                self.send_chat(NO_BLOCKS)

            if self.invalid_build_position(x, y, z):
                return False

            if self.last_blocks is None:
                self.last_blocks = []
            if BLOCK_REMOVAL_DELAY.get() > 0:
                self.last_blocks.append(((x, y, z), get_now_in_secs()))

            self.random_color()
            return can_build

        def on_block_destroy(self, x, y, z, value):
            is_trusted = (self.admin or self.god or self.user_types.moderator or
                          self.user_types.guard or self.user_types.trusted)
            if value == DESTROY_BLOCK:
                blocks = ((x, y, z),)
            elif value == SPADE_DESTROY:
                blocks = ((x, y, z), (x, y, z + 1), (x, y, z - 1))
            elif value == GRENADE_DESTROY:
                blocks = []
                for nade_x in range(x - 1, x + 2):
                    for nade_y in range(y - 1, y + 2):
                        for nade_z in range(z - 1, z + 2):
                            blocks.append((nade_x, nade_y, nade_z))

            for block in blocks:
                is_last_block_removal = False
                if self.last_blocks is not None and not is_trusted:
                    for last in self.last_blocks:
                        if block == last[0]:
                            if last[1] + BLOCK_REMOVAL_DELAY.get() < get_now_in_secs():
                                self.last_blocks.remove(last)
                                return False
                            self.last_blocks.remove(last)
                            is_last_block_removal = True
                            break
                if is_last_block_removal:
                    continue

                block_info = self.protocol.map.get_point(*block)
                if block_info[0] is True:
                    block_hls = byte_rgb_to_hls(block_info[1])
                    is_blue_block = compare_hs(block_hls, self.protocol.blue_team.hls)
                    is_green_block = compare_hs(block_hls, self.protocol.green_team.hls)
                    is_team_block = ((self.team is self.protocol.blue_team and is_blue_block) or
                                     (self.team is self.protocol.green_team and is_green_block))
                    if is_team_block and not is_trusted:
                        self.send_chat(CANT_DESTROY)
                        return False
                    if PROTECT_MAP_BLOCKS.get() and not is_blue_block and not is_green_block:
                        return False
            return connection.on_block_destroy(self, x, y, z, value)

        def on_flag_take(self):
            if self.last_spawn_time + INTEL_PICKUP_DELAY.get() > get_now_in_secs():
                return False
            return connection.on_flag_take(self)

        def on_spawn(self, pos):
            self.last_spawn_time = get_now_in_secs()
            self.last_blocks = None
            return connection.on_spawn(self, pos)

        def grenade_exploded(self, grenade):
            if DISABLE_GRENADES_AT_SPAWN.get():
                if not (self is None or self.name is None or
                        self.team is None or self.team.other is None):
                    grenade_x = int(grenade.position.x)
                    grenade_y = int(grenade.position.y)
                    spawn_x = self.team.other.spawn[0]
                    spawn_y = self.team.other.spawn[1]
                    spawn_range = self.protocol.spawn_range + 8
                    if is_in_area(grenade_x, grenade_y,
                                  spawn_x - spawn_range, spawn_y - spawn_range,
                                  spawn_x + spawn_range, spawn_y + spawn_range):
                        return False
            return connection.grenade_exploded(self, grenade)

    class PushProtocol(protocol):
        game_mode = CTF_MODE
        spawn_range = 0
        cp_protect_range = 0
        check_loop = None
        reset_intel_blue_timer = 0
        reset_intel_green_timer = 0

        def __init__(self, *arg, **kw):
            protocol.__init__(self, *arg, **kw)
            self.blue_team.hls = byte_rgb_to_hls(self.blue_team.color)
            self.blue_team.light_range = byte_middle_range(
                self.blue_team.hls[1])

            self.green_team.hls = byte_rgb_to_hls(self.green_team.color)
            self.green_team.light_range = byte_middle_range(
                self.green_team.hls[1])

        def check_intel_location(self, team, timer_val):
            if team.flag is not None:
                if team.flag.get()[2] >= 63:
                    reset_intel_position(self, team)
                    return 0
                elif team.flag.player is None:
                    timer_val += 1
                    if timer_val >= RESET_INTEL_AFTER_DROP.get() / self.check_loop.interval:
                        reset_intel_position(self, team)
                        return 0
                    return timer_val
                else:
                    return 0
            return timer_val

        def check_intel_locations(self):
            self.reset_intel_blue_timer = self.check_intel_location(
                                              self.blue_team, self.reset_intel_blue_timer)
            self.reset_intel_green_timer = self.check_intel_location(
                                              self.green_team, self.reset_intel_green_timer)

        def on_map_change(self, map):
            extensions = self.map_info.extensions
            for must_have in ('push_blue_spawn', 'push_green_spawn',
                              'push_blue_cp', 'push_green_cp'):
                if must_have not in extensions:
                    raise Exception("Missing push map metadata: %s" % must_have)

            extensions['water_damage'] = 100
            # distance from spawn center to randomly spawn in
            self.spawn_range = extensions.get('push_spawn_range', 5)
            # distance from cp where building is not allowed
            self.cp_protect_range = extensions.get('push_cp_protect_range',
                                                   DEFAULT_CP_PROTECT_RANGE.get())

            self.blue_team.spawn = extensions.get('push_blue_spawn')
            self.blue_team.cp = extensions.get('push_blue_cp')
            self.blue_team.build_area = extensions.get('push_blue_build_area')

            self.green_team.spawn = extensions.get('push_green_spawn')
            self.green_team.cp = extensions.get('push_green_cp')
            self.green_team.build_area = extensions.get('push_green_build_area')

            self.map_info.get_entity_location = get_entity_location
            self.map_info.get_spawn_location = get_spawn_location

            if self.check_loop is not None:
                self.check_loop.stop()
            self.reset_intel_blue_timer = 0
            self.reset_intel_green_timer = 0
            self.check_loop = LoopingCall(self.check_intel_locations)
            self.check_loop.start(3, now=False)

            return protocol.on_map_change(self, map)

    return PushProtocol, PushConnection



"""
/rapid [player] will put the player in rapid mode, speeding up all tools
including weapons.

RAPID_BLOCK_DELAY determines how fast block placement should be.
Lowering this value is not recommended except for local use. Ping to the server
is the effective lower limit: if this value is set to 0.1 (0.1 seconds = 100ms),
users with ping above that won't have the same advantage as the rest of the
players.

Set ALWAYS_RAPID to TRUE to automatically get rapid when you login.

.. codeauthor:: hompy
"""

from twisted.internet.reactor import callLater
from twisted.internet.task import LoopingCall
from pyspades import contained as loaders
from piqueserver.commands import command, get_player, target_player

ALWAYS_RAPID = False
RAPID_INTERVAL = 0.08
RAPID_BLOCK_DELAY = 0.26


@command('rapid', admin_only=True)
@target_player
def toggle_rapid(connection, player):
    protocol = connection.protocol

    player.rapid = rapid = not player.rapid
    player.rapid_hack_detect = not rapid
    if rapid:
        player.rapid_loop = LoopingCall(resend_tool, player)
    else:
        if player.rapid_loop and player.rapid_loop.running:
            player.rapid_loop.stop()
        player.rapid_loop = None

    message = 'now rapid' if rapid else 'no longer rapid'
    player.send_chat("You're %s" % message)
    if connection is not player and connection in protocol.players.values():
        connection.send_chat('%s is %s' % (player.name, message))
    protocol.irc_say('* %s is %s' % (player.name, message))


def resend_tool(player):
    set_tool = loaders.SetTool()
    set_tool.player_id = player.player_id
    set_tool.value = player.tool
    if player.weapon_object.shoot:
        player.protocol.broadcast_contained(set_tool)
    else:
        player.send_contained(set_tool)


def apply_script(protocol, connection, config):
    class RapidConnection(connection):
        rapid = False
        rapid_loop = None

        def on_login(self, name):
            self.rapid = ALWAYS_RAPID
            self.rapid_hack_detect = not self.rapid
            connection.on_login(self, name)

        def on_reset(self):
            if self.rapid_loop and self.rapid_loop.running:
                self.rapid_loop.stop()
            connection.on_reset(self)

        def on_disconnect(self):
            self.rapid = False
            if self.rapid_loop and self.rapid_loop.running:
                self.rapid_loop.stop()
            self.rapid_loop = None
            connection.on_disconnect(self)

        def on_block_build(self, x, y, z):
            if self.rapid:
                delay = max(0.0, RAPID_BLOCK_DELAY - self.latency / 1000.0)
                if delay > 0.0:
                    callLater(delay, resend_tool, self)
                else:
                    resend_tool(self)
            connection.on_block_build(self, x, y, z)

        def on_grenade_thrown(self, grenade):
            if self.rapid:
                resend_tool(self)
            connection.on_grenade_thrown(self, grenade)

        def on_shoot_set(self, fire):
            if self.rapid and self.rapid_loop:
                if not self.rapid_loop.running and fire:
                    self.rapid_loop.start(RAPID_INTERVAL)
                elif self.rapid_loop.running and not fire:
                    self.rapid_loop.stop()
            connection.on_shoot_set(self, fire)

    return protocol, RapidConnection



from typing import Tuple
from pyspades.constants import (
    BLUE_FLAG, GREEN_FLAG,
    CTF_MODE,
    BLUE_BASE, GREEN_BASE,
)
from pyspades.entities import Flag, Base
from pyspades.protocol import BaseProtocol


class Team:
    score = None
    flag = None
    base = None
    other = None
    protocol = None
    name = None
    kills = None

    def __init__(self, team_id: int, name: str, color: Tuple[int, int, int],
                 spectator: bool, protocol: BaseProtocol) -> None:
        self.id = team_id
        self.name = name
        self.protocol = protocol
        self.color = color
        self.spectator = spectator

    def get_players(self) -> None:
        for player in self.protocol.players.values():
            if player.team is self:
                yield player

    def count(self) -> int:
        count = 0
        for player in self.protocol.players.values():
            if player.team is self:
                count += 1
        return count

    def initialize(self) -> None:
        if self.spectator:
            return
        self.score = 0
        self.kills = 0
        if self.protocol.game_mode == CTF_MODE:
            self.set_flag()
            self.set_base()

    def set_flag(self) -> Flag:
        entity_id = [BLUE_FLAG, GREEN_FLAG][self.id]
        if self.flag is None:
            self.flag = Flag(entity_id, self.protocol)
            self.flag.team = self
            self.protocol.entities.append(self.flag)
        location = self.get_entity_location(entity_id)
        returned = self.protocol.on_flag_spawn(location[0], location[1],
                                               location[2], self.flag, entity_id)
        if returned is not None:
            location = returned
        self.flag.set(*location)
        self.flag.player = None
        return self.flag

    def set_base(self) -> Base:
        entity_id = [BLUE_BASE, GREEN_BASE][self.id]
        if self.base is None:
            self.base = Base(entity_id, self.protocol)
            self.base.team = self
            self.protocol.entities.append(self.base)
        location = self.get_entity_location(entity_id)
        returned = self.protocol.on_base_spawn(location[0], location[1],
                                               location[2], self.base, entity_id)
        if returned is not None:
            location = returned
        self.base.set(*location)
        return self.base

    def get_entity_location(self, entity_id: int) -> Tuple[int, int, int]:
        return self.get_random_location(True)

    def get_random_location(self, force_land: bool = False) -> Tuple[int, int, int]:
        x_offset = self.id * 384
        return self.protocol.get_random_location(force_land, (
            x_offset, 128, 128 + x_offset, 384))

    def get_entities(self):
        for item in self.protocol.entities:
            if item.team is self:
                yield item

    def __repr__(self):
        return "{}(id={}, name={}, color={}, spectator={}, protocol)".format(
            self.__class__.__name__, self.id,
            self.name, self.color, self.spectator
        )



"""
Blocks built by players are twice as hard to break.

* You can remove your own blocks as if they weren't strong.
* Dirt-colored or buried blocks (those that turn into dirt) become normal blocks.

.. codeauthor:: hompy
"""

from collections import namedtuple
from pyspades.contained import BlockAction, SetColor
from pyspades.common import make_color
from pyspades.color import rgb_distance
from pyspades.constants import *

DIRT_COLOR = (71, 48, 35)

StrongBlock = namedtuple('StrongBlock', 'color owner')


def rebuild_block(player, x, y, z, color):
    set_color = SetColor()
    set_color.value = make_color(*color)
    set_color.player_id = 32
    block_action = BlockAction()
    block_action.player_id = 32
    block_action.x = x
    block_action.y = y
    block_action.z = z
    block_action.value = DESTROY_BLOCK
    player.send_contained(block_action)
    block_action.value = BUILD_BLOCK
    player.send_contained(set_color)
    player.send_contained(block_action)


def check_if_buried(protocol, x, y, z):
    if not protocol.map.is_surface(x, y, z):
        protocol.strong_blocks.pop((x, y, z), None)


def bury_adjacent(protocol, x, y, z):
    check_if_buried(protocol, x, y, z - 1)
    check_if_buried(protocol, x, y, z + 1)
    check_if_buried(protocol, x - 1, y, z)
    check_if_buried(protocol, x + 1, y, z)
    check_if_buried(protocol, x, y - 1, z)
    check_if_buried(protocol, x, y + 1, z)


def is_color_dirt(color):
    return rgb_distance(color, DIRT_COLOR) < 30


def apply_script(protocol, connection, config):
    class StrongBlockConnection(connection):

        def on_disconnect(self):
            strong_blocks = self.protocol.strong_blocks
            for xyz, strong_block in strong_blocks.items():
                if strong_block.owner is self:
                    strong_blocks[xyz] = strong_block._replace(owner=None)
            connection.on_disconnect(self)

        def on_block_build(self, x, y, z):
            if not is_color_dirt(self.color):
                strong_block = StrongBlock(self.color, self)
                self.protocol.strong_blocks[(x, y, z)] = strong_block
                bury_adjacent(self.protocol, x, y, z)
            connection.on_block_build(self, x, y, z)

        def on_line_build(self, points):
            if not is_color_dirt(self.color):
                strong_block = StrongBlock(self.color, self)
                for xyz in points:
                    self.protocol.strong_blocks[xyz] = strong_block
                    bury_adjacent(self.protocol, *xyz)
            connection.on_line_build(self, points)

        def on_block_destroy(self, x, y, z, value):
            can_destroy = connection.on_block_destroy(self, x, y, z, value)
            if can_destroy != False:
                strong_block = self.protocol.strong_blocks.pop((x, y, z), None)
                if strong_block is not None:
                    # block is a strong block
                    if strong_block.owner is not self:
                        # tough for everyone but the player who built it
                        rebuild_block(self, x, y, z, strong_block.color)
                        return False
            return can_destroy

    class StrongBlockProtocol(protocol):
        strong_blocks = None

        def on_map_change(self, map):
            self.strong_blocks = {}
            protocol.on_map_change(self, map)

    return StrongBlockProtocol, StrongBlockConnection



"""
Gives a specified amount of medkits on spawn

Commands
^^^^^^^^

* ``/medkit or /m`` utilizes available medkits to heal

Options
^^^^^^^

.. code-block:: guess

   [medkit]
   medkits = 1 # no. of medkits
   medkit_heal_amount = 40 # how much hp. it gives

.. codeauthor:: Booboorocks998 & mat^2
"""

from piqueserver.commands import command
from piqueserver.config import config
from pyspades.constants import FALL_KILL

medkit_config = config.section("medkit")
default_medkits = medkit_config.option("medkits", 1)
medkit_heal_amount = medkit_config.option("medkit_heal_amount", 40)

@command('medkit', 'm')
def medkit(connection):
    if connection.medkits and connection.hp < 100:
        connection.set_hp(connection.hp + connection.protocol.heal_amount,
                          kill_type=FALL_KILL)
        connection.medkits -= 1
        connection.send_chat('You have been healed')
    else:
        connection.send_chat("You don't have any medkits or have full health!")


def apply_script(protocol, connection, config):
    class MedkitConnection(connection):
        medkits = 0

        def on_spawn(self, pos):
            self.medkits = default_medkits.get()
            self.send_chat('You have %s medkit!' % self.medkits)
            return connection.on_spawn(self, pos)

    class MedkitProtocol(protocol):
        heal_amount = medkit_heal_amount.get()

    return MedkitProtocol, MedkitConnection



"""
Lets you change the color of the block you are looking at.
With block tool selected, pick a color, then hold down sneak key
(<V> by default) to paint.

Commands
^^^^^^^^

* ``/paint <player>`` enables painting mode for that player.

.. codeauthor:: hompy
"""

from pyspades.contained import BlockAction
from pyspades.common import Vertex3
from pyspades.constants import *
from piqueserver.commands import command, admin, get_player, target_player

PAINT_RAY_LENGTH = 32.0


@command(admin_only=True)
@target_player
def paint(connection, player):
    protocol = connection.protocol

    player.painting = not player.painting

    message = 'now painting' if player.painting else 'no longer painting'
    player.send_chat("You're %s" % message)
    if connection is not player and connection in protocol.players.values():
        connection.send_chat('%s is %s' % (player.name, message))
    protocol.irc_say('* %s is %s' % (player.name, message))


def paint_block(protocol, player, x, y, z, color):
    if x < 0 or y < 0 or z < 0 or x >= 512 or y >= 512 or z >= 62:
        return False
    if protocol.map.get_color(x, y, z) == color:
        return False
    protocol.map.set_point(x, y, z, color)
    block_action = BlockAction()
    block_action.x = x
    block_action.y = y
    block_action.z = z
    block_action.player_id = player.player_id
    block_action.value = DESTROY_BLOCK
    protocol.broadcast_contained(block_action, save=True)
    block_action.value = BUILD_BLOCK
    protocol.broadcast_contained(block_action, save=True)
    return True


def paint_ray(player):
    if player.tool != BLOCK_TOOL:
        return
    location = player.world_object.cast_ray(PAINT_RAY_LENGTH)
    if location:
        x, y, z = location
        if player.on_block_build_attempt(x, y, z) == False:
            return
        paint_block(player.protocol, player, x, y, z, player.color)


def apply_script(protocol, connection, config):
    class PaintConnection(connection):
        painting = False

        def on_reset(self):
            self.painting = False
            connection.on_reset(self)

        def on_position_update(self):
            if self.painting and self.world_object.sneak:
                paint_ray(self)
            connection.on_position_update(self)

        def on_orientation_update(self, x, y, z):
            if self.painting and self.world_object.sneak:
                paint_ray(self)
            connection.on_orientation_update(self, x, y, z)

        def on_animation_update(self, jump, crouch, sneak, sprint):
            if self.painting and sneak:
                paint_ray(self)
            return connection.on_animation_update(
                self, jump, crouch, sneak, sprint)

    return protocol, PaintConnection



from piqueserver.commands import (command, get_player, join_arguments,
                                  player_only)


@command()
@player_only
def login(connection, password):
    """
    Log in if you're staff or a trusted member of this server
    /login <password>
    You will be kicked if a wrong password is given 3 times in a row
    """
    for user_type, passwords in connection.protocol.passwords.items():
        if password in passwords:
            if user_type in connection.user_types:
                return "You're already logged in as %s" % user_type
            return connection.on_user_login(user_type, True)
    if connection.login_retries is None:
        connection.login_retries = connection.protocol.login_retries - 1
    else:
        connection.login_retries -= 1
    if not connection.login_retries:
        connection.kick('Ran out of login attempts')
        return
    return 'Invalid password - you have %s tries left' % (
        connection.login_retries)


@command()
def pm(connection, value, *arg):
    """
    Send a private message to a given player
    /pm <player> <message>
    """
    player = get_player(connection.protocol, value)
    message = join_arguments(arg)
    if len(message) == 0:
        return "Please specify your message"
    player.send_chat('PM from %s: %s' % (connection.name, message))
    return 'PM sent to %s' % player.name


@command('admin')
def to_admin(connection, *arg):
    """
    Send a message to all admins currently online
    /admin <message>
    """
    protocol = connection.protocol
    message = join_arguments(arg)
    if not message:
        return "Enter a message you want to send, like /admin I'm stuck"
    prefix = '(TO ADMINS)'
    irc_relay = protocol.irc_relay
    if irc_relay:
        if irc_relay.factory.bot and irc_relay.factory.bot.colors:
            prefix = '\x0304' + prefix + '\x0f'
        irc_relay.send(prefix + ' <%s> %s' % (connection.name, message))
    for player in protocol.players.values():
        if player.admin and player is not connection:
            player.send_chat('To ADMINS from %s: %s' %
                             (connection.name, message))
    return 'Message sent to all admins currently online'



import math
from typing import List, Tuple, Optional, Union

from twisted.internet import reactor
from twisted.logger import Logger

from piqueserver import commands
from piqueserver.release import format_release
import pyspades
from pyspades.constants import (ERROR_BANNED, DESTROY_BLOCK, SPADE_DESTROY,
                                GRENADE_DESTROY, ERROR_KICKED, BLOCK_TOOL)
from pyspades.server import ServerConnection
from pyspades.common import encode, escape_control_codes, prettify_timespan, Vertex3
from pyspades.world import Character
from pyspades.types import AttributeSet

# TODO: move these where they belong
from pyspades.team import Team
from pyspades.world import Grenade
CHAT_WINDOW_SIZE = 5
CHAT_PER_SECOND = 0.5

HookValue = Optional[bool]

log = Logger()


class FeatureConnection(ServerConnection):
    printable_name = None
    admin = False
    last_switch = None
    mute = False
    deaf = False
    login_retries = None
    god = False
    god_build = False
    fly = False
    invisible = False
    building = True
    killing = True
    streak = 0
    best_streak = 0
    last_chat = None
    chat_time = 0
    chat_count = 0
    user_types = None
    rights = None
    can_complete_line_build = True

    def on_connect(self) -> None:
        protocol = self.protocol
        client_ip = self.address[0]

        if client_ip in self.protocol.bans:
            name, reason, timestamp = self.protocol.bans[client_ip]

            if timestamp is not None and reactor.seconds() >= timestamp:
                protocol.remove_ban(client_ip)
                protocol.save_bans()
            else:
                log.info('banned user {} ({}) attempted to join'.format(name,
                                                                        client_ip))
                self.disconnect(ERROR_BANNED)
                return

        manager = self.protocol.ban_manager

        if manager is not None:
            reason = manager.get_ban(client_ip)
            if reason is not None:
                log.info(('federated banned user (%s) attempted to join, '
                          'banned for %r') % (client_ip, reason))
                self.disconnect(ERROR_BANNED)
                return

        ServerConnection.on_connect(self)

    def on_join(self) -> None:
        if self.protocol.motd is not None:
            self.send_lines(self.protocol.motd)

    def on_login(self, name: str) -> None:
        self.printable_name = escape_control_codes(name)
        if len(self.printable_name) > 15:
            self.kick(silent=True)
        log.info('{name} (IP {ip}, ID {pid}) entered the game!',
                 name=self.printable_name,
                 ip=self.address[0], pid=self.player_id)
        self.protocol.irc_say('* %s (IP %s, ID %s) entered the game!' %
                              (self.name, self.address[0], self.player_id))
        if self.user_types is None:
            self.user_types = AttributeSet()
            self.rights = AttributeSet()
            if self.protocol.everyone_is_admin:
                self.on_user_login('admin', False)

    def get_spawn_location(self) -> Tuple[int, int, int]:
        get_location = self.protocol.map_info.get_spawn_location
        if get_location is not None:
            result = get_location(self)
            if result is not None:
                return result
        return ServerConnection.get_spawn_location(self)

    def on_disconnect(self) -> None:
        if self.name is not None:
            log.info('{name} disconnected!', name=self.printable_name)
            self.protocol.irc_say('* %s (IP %s) disconnected' %
                                  (self.name, self.address[0]))
            self.protocol.player_memory.append((self.name, self.address[0]))
        else:
            log.info('{ip} disconnected', ip=self.address[0])
        ServerConnection.on_disconnect(self)

    def on_command(self, command: str, parameters: List[str]) -> None:
        result = commands.handle_command(self, command, parameters)

        if result:
            for i in reversed(result.split("\n")):
                self.send_chat(i)

    def _can_build(self) -> bool:
        if not self.building:
            return False
        if not self.god and not self.protocol.building:
            return False

        return True

    def on_block_build_attempt(self, x: int, y: int, z: int) -> bool:
        return self._can_build()

    def on_line_build_attempt(self, points) -> bool:
        return self._can_build()

    def on_line_build(self, points) -> None:
        if self.god:
            self.refill()
        if self.god_build:
            if self.protocol.god_blocks is None:
                self.protocol.god_blocks = set()
            self.protocol.god_blocks.update(points)
        elif self.protocol.user_blocks is not None:
            self.protocol.user_blocks.update(points)

    def on_block_build(self, x: int, y: int, z: int) -> None:
        if self.god:
            self.refill()
        if self.god_build:
            if self.protocol.god_blocks is None:
                self.protocol.god_blocks = set()
            self.protocol.god_blocks.add((x, y, z))
        elif self.protocol.user_blocks is not None:
            self.protocol.user_blocks.add((x, y, z))

    def on_block_destroy(self, x: int, y: int, z: int, mode: int) -> bool:
        map_on_block_destroy = self.protocol.map_info.on_block_destroy
        if map_on_block_destroy is not None:
            result = map_on_block_destroy(self, x, y, z, mode)
            if result == False:
                return result
        if not self.building:
            return False
        if not self.god:
            if not self.protocol.building:
                return False
            is_indestructable = self.protocol.is_indestructable
            if mode == DESTROY_BLOCK:
                if is_indestructable(x, y, z):
                    return False
            elif mode == SPADE_DESTROY:
                if (is_indestructable(x, y, z) or
                        is_indestructable(x, y, z + 1) or
                        is_indestructable(x, y, z - 1)):
                    return False
            elif mode == GRENADE_DESTROY:
                for nade_x in range(x - 1, x + 2):
                    for nade_y in range(y - 1, y + 2):
                        for nade_z in range(z - 1, z + 2):
                            if is_indestructable(nade_x, nade_y, nade_z):
                                return False

    def on_block_removed(self, x: int, y: int, z: int) -> None:
        if self.protocol.user_blocks is not None:
            self.protocol.user_blocks.discard((x, y, z))
        if self.protocol.god_blocks is not None:
            self.protocol.god_blocks.discard((x, y, z))

    def on_hit(self, hit_amount: float, player: 'FeatureConnection',
               _type: int, grenade: Grenade) -> HookValue:
        if not self.protocol.killing:
            self.send_chat(
                "You can't kill anyone right now! Damage is turned OFF")
            return False
        if not self.killing:
            self.send_chat("%s. You can't kill anyone." % player.name)
            return False
        elif player.god:
            if not player.invisible:
                self.send_chat("You can't hurt %s! That player is in "
                               "*god mode*" % player.name)
            return False
        if self.god:
            self.protocol.send_chat('%s, killing in god mode is forbidden!' %
                                    self.name, irc=True)
            self.protocol.send_chat('%s returned to being a mere human.' %
                                    self.name, irc=True)
            self.god = False
            self.god_build = False

    def on_kill(self, killer: Optional['FeatureConnection'], _type: int,
                grenade: None) -> None:
        self.streak = 0
        if killer is None or self.team is killer.team:
            return
        if not grenade or grenade.name == 'grenade':
            # doesn't give streak kills on airstrikes (or other types of
            # explosions)
            killer.streak += 1
            killer.best_streak = max(killer.streak, killer.best_streak)
        killer.team.kills += 1

    def on_reset(self) -> None:
        self.streak = 0
        self.best_streak = 0

    def on_animation_update(self, jump: bool, crouch: bool, sneak: bool,
                            sprint: bool) -> Tuple[bool, bool, bool, bool]:
        if self.fly and crouch and self.world_object.velocity.z != 0.0:
            jump = True
        return jump, crouch, sneak, sprint

    def on_fall(self, damage: int) -> HookValue:
        if self.god:
            return False
        if not self.protocol.fall_damage:
            return False

    def on_grenade(self, time_left: float) -> None:
        if self.god:
            self.refill()

    def on_team_join(self, team: 'FeatureTeam') -> HookValue:
        if self.team is not None:
            if self.protocol.teamswitch_interval:
                teamswitch_interval = self.protocol.teamswitch_interval
                teamswitch_allowed = self.protocol.teamswitch_allowed
                if not teamswitch_allowed:
                    self.send_chat('Switching teams is not allowed')
                    return False
                if (self.last_switch is not None and
                        reactor.seconds() - self.last_switch < teamswitch_interval):
                    self.send_chat(
                        'You must wait before switching teams again')
                    return False
        if team.locked:
            self.send_chat('Team is locked')
            if not team.spectator and not team.other.locked:
                return team.other
            return False
        balanced_teams = self.protocol.balanced_teams
        if balanced_teams and not team.spectator:
            other_team = team.other
            if other_team.count() < team.count() + 1 - balanced_teams:
                if other_team.locked:
                    return False
                self.send_chat('Team is full, moved to %s' % other_team.name)
                return other_team
        self.last_switch = reactor.seconds()

    def on_chat(self, value: str, global_message: bool) -> Union[str, bool]:
        """
        notifies when the server receives a chat message

        return False to block sending the message
        """
        message = '<{}> {}'.format(self.name, value)

        if self.mute:
            message = '(MUTED) {}'.format(message)
            self.send_chat('(Chat not sent - you are muted)')
            return False

        if global_message:
            if self.protocol.global_chat:
                # forward message to IRC
                self.protocol.irc_say(message)
            else:
                self.send_chat('(Chat not sent - global chat disabled)')
                return False

        # antispam:
        current_time = reactor.seconds()
        if self.last_chat is None:
            self.last_chat = current_time

        else:
            self.chat_time += current_time - self.last_chat

            if self.chat_count > CHAT_WINDOW_SIZE:
                if self.chat_count / self.chat_time > CHAT_PER_SECOND:
                    self.mute = True
                    self.protocol.send_chat(
                        '%s has been muted for excessive spam' % (
                            self.name),
                        irc=True)

                # reset if CHAT_WINDOW_SIZE messages were sent and not
                # determined to be spam
                self.chat_time = 0
                self.chat_count = 0
            else:
                self.chat_count += 1
            self.last_chat = current_time

        log.info("<{name}> {message}", name=escape_control_codes(
            self.name), message=escape_control_codes(value))

        return value

    def kick(self, reason=None, silent=False):
        if not silent:
            if reason is not None:
                message = '{} was kicked: {}'.format(self.name, reason)
            else:
                message = '%s was kicked' % self.name
            self.protocol.send_chat(message, irc=True)
            log.info(message)
        # FIXME: Client should handle disconnect events the same way in both
        # main and initial loading network loops
        self.disconnect(ERROR_KICKED)

    def ban(self, reason=None, duration=None):
        reason = ': ' + reason if reason is not None else ''
        duration = duration or None
        if duration is None:
            message = '{} permabanned{}'.format(self.name, reason)
        else:
            message = '{} banned for {}{}'.format(self.name,
                                                  prettify_timespan(duration), reason)
        if self.protocol.on_ban_attempt(self, reason, duration):
            self.protocol.send_chat(message, irc=True)
            self.protocol.on_ban(self, reason, duration)
            if self.address[0] == "127.0.0.1":
                self.protocol.send_chat("Ban ignored: localhost")
            else:
                self.protocol.add_ban(self.address[0], reason, duration,
                                      self.name)

    def send_lines(self, lines: List[str]) -> None:
        current_time = 0
        for line in lines:
            reactor.callLater(current_time, self.send_chat, line)
            current_time += 2

    def on_hack_attempt(self, reason):
        log.warn('Hack attempt detected from {}: {}'.format(self.printable_name,
                                                            reason))
        self.kick(reason)

    def on_user_login(self, user_type, verbose=True):
        log.info("'{username}' logged in as {user_type}", username=self.name,
                 user_type=user_type)

        if user_type == 'admin':
            self.admin = True
            self.speedhack_detect = False

        # notify of new release to admin on /login
        new_release = self.protocol.new_release
        if user_type == 'admin' and new_release:
            self.send_chat("!" * 30)
            self.send_chat(format_release(new_release))
            self.send_chat("!" * 30)

        self.user_types.add(user_type)
        rights = set(commands.get_rights(user_type))
        self.rights.update(rights)
        if verbose:
            message = ' logged in as %s' % (user_type)
            self.send_chat('You' + message)
            self.protocol.irc_say("* " + self.name + message)

    def timed_out(self):
        if self.name is not None:
            log.info('%s timed out' % self.printable_name)
        ServerConnection.timed_out(self)



"""
Tug of War game mode, where you must progressively capture the enemy CPs in a
straight line to win.

Maintainer: mat^2
"""

import random
import math
from pyspades.constants import TC_MODE
from pyspades.server import Territory

CP_COUNT = 6
CP_EXTRA_COUNT = CP_COUNT + 2  # PLUS last 'spawn'
ANGLE = 65
START_ANGLE = math.radians(-ANGLE)
END_ANGLE = math.radians(ANGLE)
DELTA_ANGLE = math.radians(30)
FIX_ANGLE = math.radians(4)

HELP = [
    "In Tug of War, you capture your opponents' front CP to advance."
]


class TugTerritory(Territory):
    disabled = True

    def add_player(self, player):
        if self.disabled:
            return
        Territory.add_player(self, player)

    def enable(self):
        self.disabled = False

    def disable(self):
        for player in self.players.copy():
            self.remove_player(player)
        self.disabled = True
        self.progress = float(self.team.id)


def get_index(value):
    if value < 0:
        raise IndexError()
    return value


def random_up_down(value):
    value /= 2
    return random.uniform(-value, value)


def limit_angle(value):
    return min(END_ANGLE, max(START_ANGLE, value))


def limit_dimension(value):
    return min(511, max(0, value))


def get_point(x, y, magnitude, angle):
    return (limit_dimension(x + math.cos(angle) * magnitude),
            limit_dimension(y + math.sin(angle) * magnitude))


def apply_script(protocol, connection, config):
    class TugConnection(connection):

        def get_spawn_location(self):
            if self.team.spawn_cp is None:
                base = self.team.last_spawn
            else:
                base = self.team.spawn_cp
            return base.get_spawn_location()

        def on_spawn(self, pos):
            for line in HELP:
                self.send_chat(line)
            return connection.on_spawn(self, pos)

    class TugProtocol(protocol):
        game_mode = TC_MODE

        def get_cp_entities(self):
            # generate positions

            map = self.map
            blue_cp = []
            green_cp = []

            magnitude = 10
            angle = random.uniform(START_ANGLE, END_ANGLE)
            x, y = (0, random.randrange(64, 512 - 64))

            points = []

            square_1 = range(128)
            square_2 = range(512 - 128, 512)

            while True:
                top = int(y) in square_1
                bottom = int(y) in square_2
                if top:
                    angle = limit_angle(angle + FIX_ANGLE)
                elif bottom:
                    angle = limit_angle(angle - FIX_ANGLE)
                else:
                    angle = limit_angle(angle + random_up_down(DELTA_ANGLE))
                magnitude += random_up_down(2)
                magnitude = min(15, max(5, magnitude))
                x2, y2 = get_point(x, y, magnitude, angle)
                if x2 >= 511:
                    break
                x, y = x2, y2
                points.append((int(x), int(y)))

            move = 512 / CP_EXTRA_COUNT
            offset = move / 2

            for i in range(CP_EXTRA_COUNT):
                index = 0
                while True:
                    p_x, p_y = points[index]
                    index += 1
                    if p_x >= offset:
                        break
                if i < CP_EXTRA_COUNT / 2:
                    blue_cp.append((p_x, p_y))
                else:
                    green_cp.append((p_x, p_y))
                offset += move

            # make entities

            index = 0
            entities = []

            for i, (x, y) in enumerate(blue_cp):
                entity = TugTerritory(index, self, *(x, y, map.get_z(x, y)))
                entity.team = self.blue_team
                if i == 0:
                    self.blue_team.last_spawn = entity
                    entity.id = -1
                else:
                    entities.append(entity)
                    index += 1

            self.blue_team.cp = entities[-1]
            self.blue_team.cp.disabled = False
            self.blue_team.spawn_cp = entities[-2]

            for i, (x, y) in enumerate(green_cp):
                entity = TugTerritory(index, self, *(x, y, map.get_z(x, y)))
                entity.team = self.green_team
                if i == len(green_cp) - 1:
                    self.green_team.last_spawn = entity
                    entity.id = index
                else:
                    entities.append(entity)
                    index += 1

            self.green_team.cp = entities[-CP_COUNT // 2]
            self.green_team.cp.disabled = False
            self.green_team.spawn_cp = entities[-CP_COUNT // 2 + 1]

            return entities

        def on_cp_capture(self, territory):
            team = territory.team
            if team.id:
                move = -1
            else:
                move = 1
            for team in [self.blue_team, self.green_team]:
                try:
                    team.cp = self.entities[get_index(team.cp.id + move)]
                    team.cp.enable()
                except IndexError:
                    pass
                try:
                    team.spawn_cp = self.entities[get_index(
                        team.spawn_cp.id + move)]
                except IndexError:
                    team.spawn_cp = team.last_spawn
            cp = (self.blue_team.cp, self.green_team.cp)
            for entity in self.entities:
                if not entity.disabled and entity not in cp:
                    entity.disable()

    return TugProtocol, TugConnection



"""
Smartnade script

.. warning::
   This script conflicts with `minefield` script.

.. codeauthor:: ?
"""


SMARTNADE_DELAY = 0.5


def apply_script(protocol, connection, config):
    class SmartNadeProtocol(protocol):

        def on_world_update(self):
            for player in list(self.players.values()):
                for nade in player.smart_nades:
                    if nade.fuse > SMARTNADE_DELAY:
                        for enemy in player.team.other.get_players():
                            if nade.get_damage(
                                    enemy.world_object.position) != 0:
                                nade.fuse = min(nade.fuse, SMARTNADE_DELAY)
            return protocol.on_world_update(self)

    class SmartNadeConnection(connection):

        def __init__(self, *arg, **kw):
            connection.__init__(self, *arg, **kw)
            self.smart_nades = []

        def on_grenade_thrown(self, grenade):
            self.smart_nades.append(grenade)
            return connection.on_grenade_thrown(self, grenade)

        def grenade_exploded(self, grenade):
            self.smart_nades.remove(grenade)
            return connection.grenade_exploded(self, grenade)

    return SmartNadeProtocol, SmartNadeConnection



"""
Links any two teammates so that they explode if they stray away from each other.
An incredibly unnerving method of encouraging teamplay.

Enable by doing /runningman or by setting ENABLED_AT_START = True
Linking is automatic at spawn and tries to find new partners every two deaths.
Use /unlink [player] to toggle yourself or someone from being linked.

May not work well with squads.

.. codeauthor:: hompy
"""

from random import choice
from twisted.internet.reactor import seconds
from pyspades.world import Grenade
from pyspades.contained import GrenadePacket
from pyspades.collision import distance_3d_vector
from pyspades.constants import *
from pyspades.common import Vertex3
from piqueserver.commands import (command, admin, get_player, target_player)

ENABLED_AT_START = False

LINK_DISTANCE = 40.0
LINK_WARNING_DISTANCE = LINK_DISTANCE * 0.65
MAX_LINK_DEATHS = 2

S_ENABLED = 'Running Man mode ENABLED'
S_DISABLED = 'Running Man mode DISABLED'
S_NOT_ENABLED = 'Running Man mode is not enabled'
S_LINK_BREAK = "You strayed too far from {player}... don't go losing your " \
    "head over it"
S_LINK_WARNING = 'WARNING! Get back to {player} or you will both die!'
S_LINKED = "You've been linked to {player}. Stay close to your partner or die!"
S_FREE = "You're free to roam around... for now"
S_LINKABLE = "{player}'s collar hums back to life"
S_LINKABLE_SELF = "Your collar hums back to life"
S_UNLINKABLE = '{player} is free as a bird'
S_UNLINKABLE_SELF = "You're free as a bird"
S_UNLINK_ALL = 'All players unlinked!'
S_FLAG_CAPTURED = 'The intel capture has unlinked everyone in the {team} team!'


@command('runningman', admin_only=True)
def running_man(connection):
    protocol = connection.protocol
    protocol.running_man = not protocol.running_man
    if not protocol.running_man:
        protocol.drop_all_links()
    message = S_ENABLED if protocol.running_man else S_DISABLED
    protocol.send_chat(message, irc=True)


@command(admin_only=True)
def relink(connection):
    if not connection.protocol.running_man:
        return S_NOT_ENABLED
    connection.protocol.drop_all_links()
    connection.protocol.send_chat(S_UNLINK_ALL, irc=True)


@command(admin_only=True)
@target_player
def unlink(connection, player):
    protocol = connection.protocol
    if not protocol.running_man:
        return S_NOT_ENABLED

    player.drop_link()
    player.linkable = not player.linkable
    player.send_chat(S_LINKABLE_SELF if player.linkable else S_UNLINKABLE_SELF)
    message = S_LINKABLE if player.linkable else S_UNLINKABLE
    message.format(player=player.name)
    if connection is not player:
        return message


def apply_script(protocol, connection, config):
    class RunningManConnection(connection):
        link = None
        linkable = True
        last_warning = None
        link_deaths = None

        def on_position_update(self):
            if self.protocol.running_man:
                if self.link is not None and self.link.hp is not None and self.link.hp > 0:
                    dist = distance_3d_vector(self.world_object.position,
                                              self.link.world_object.position)
                    if dist > LINK_DISTANCE:
                        self.grenade_suicide()
                        self.link_deaths += 1
                        self.link.grenade_suicide()
                        self.link.link_deaths += 1

                        message = S_LINK_BREAK.format(player=self.link.name)
                        self.send_chat(message)
                        message = S_LINK_BREAK.format(player=self.name)
                        self.link.send_chat(message)
                    elif (dist > LINK_WARNING_DISTANCE and
                          (self.last_warning is None or
                           seconds() - self.last_warning > 2.0)):

                        self.last_warning = seconds()
                        self.link.last_warning = seconds()

                        message = S_LINK_WARNING.format(player=self.link.name)
                        self.send_chat(message)
                        message = S_LINK_WARNING.format(player=self.name)
                        self.link.send_chat(message)
            connection.on_position_update(self)

        def on_spawn(self, pos):
            if self.protocol.running_man:
                if (self.link is None or
                        self.link_deaths >= MAX_LINK_DEATHS):
                    self.get_new_link()
                if self.link is not None and self.link.hp is not None and self.link.hp > 0:
                    self.set_location_safe(
                        self.link.world_object.position.get())
            connection.on_spawn(self, pos)

        def on_team_changed(self, old_team):
            if self.protocol.running_man:
                self.drop_link()
            return connection.on_team_changed(self, old_team)

        def on_flag_capture(self):
            if self.protocol.running_man:
                for player in self.team.get_players():
                    player.drop_link(no_message=True)
                message = S_FLAG_CAPTURED.format(team=self.team.name)
                self.protocol.send_chat(message, global_message=None)
            connection.on_flag_capture(self)

        def on_reset(self):
            if self.protocol.running_man:
                self.drop_link()
            connection.on_reset(self)

        def can_be_linked_to(self, player):
            if self is player or self.link is player:
                return False
            if player.link is not None and player.link_deaths < MAX_LINK_DEATHS:
                return False
            return True

        def get_new_link(self):
            available = list(filter(self.can_be_linked_to,
                                    self.team.get_players()))
            if not available:
                return
            self.drop_link(force_message=True)
            self.link = choice(available)
            self.link_deaths = 0
            self.link.drop_link(force_message=True)
            self.link.link = self
            self.link.link_deaths = 0

            message = S_LINKED.format(player=self.link.name)
            self.send_chat(message)
            message = S_LINKED.format(player=self.name)
            self.link.send_chat(message)

        def drop_link(self, force_message=False, no_message=False):
            if self.link is None:
                return
            if (self.link.hp is not None and (self.link.hp > 0 or force_message)) and not no_message:
                self.link.send_chat(S_FREE)
            self.link.link = None
            self.link = None

        def grenade_suicide(self):
            protocol = self.protocol
            position = self.world_object.position
            protocol.world.create_object(Grenade, 0.0, position, None,
                                         Vertex3(), self.grenade_exploded)
            grenade_packet = GrenadePacket()
            grenade_packet.value = 0.0
            grenade_packet.player_id = self.player_id
            grenade_packet.position = position.get()
            grenade_packet.velocity = (0.0, 0.0, 0.0)
            protocol.broadcast_contained(grenade_packet)
            self.kill(kill_type=GRENADE_KILL)

    class RunningManProtocol(protocol):
        running_man = ENABLED_AT_START

        def drop_all_links(self):
            for player in list(self.players.values()):
                player.drop_link(no_message=True)

    return RunningManProtocol, RunningManConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

# FIXME: This should be probably moved into it's own URL thing and put into
# some module with a name describing it's insignificance

from ipaddress import IPv4Address
def make_server_identifier(ip: IPv4Address, port: int = 32887) -> str:
    # ip should be an IPv4Address object
    a, b, c, d = ip.exploded.split('.')
    a = int(a)
    b = int(b)
    c = int(c)
    d = int(d)
    return 'aos://{}:{}'.format(a | (b << 8) | (c << 16) | (d << 24), port)


def get_server_details(value):
    if not value.startswith('aos://'):
        raise ValueError('invalid server identifier')
    splitted = value[6:].split(':')
    if len(splitted) == 1:
        host = int(splitted[0])
        port = 32887
    else:
        host, port = splitted
        host = int(host)
        port = int(port)
    a = host & 0xFF
    b = (host & 0xFF00) >> 8
    c = (host & 0xFF0000) >> 16
    d = (host & 0xFF000000) >> 24
    return ('{}.{}.{}.{}'.format(a, b, c, d), port)



"""
Prevents spade-nade bug

version 2(2017.12.23)

.. codeauthor:: kmsi<kmsiapps@gmail.com>
"""

from pyspades.constants import SPADE_TOOL

def apply_script(protocol, connection, config):
    class SpadenadeConnection(connection):
        def on_secondary_fire_set(self, secondary):
            self.was_spade = (self.tool == SPADE_TOOL)

        def on_grenade(self, time_left):
            if(self.world_object.secondary_fire and self.was_spade):
                self.send_chat('Spade-Grenade bug is blocked.')
                self.was_spade = False
                return False
            else:
                self.was_spade = False
                return connection.on_grenade(self, time_left)

    return protocol, SpadenadeConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from pyspades.common cimport allocate_memory

cdef tuple make_color_tuple(int color):
    cdef int r, g, b, a
    b = color & 0xFF
    g = (color & 0xFF00) >> 8
    r = (color & 0xFF0000) >> 16
    a = (((color & 0xFF000000) >> 24) / 128.0) * 255
    return (r, g, b)

cpdef inline int make_color(int r, int g, int b, int a = 255):
    return b | (g << 8) | (r << 16) | (<int>((a / 255.0) * 128) << 24)

import time
import random

cdef class Generator:
    cdef MapGenerator * generator
    cdef public:
        bint done

    def __init__(self, VXLData data):
        self.done = False
        self.generator = create_map_generator(data.map)

    def get_data(self, int columns = 2):
        if self.done:
            return None
        value = get_generator_data(self.generator, columns)
        if not value:
            self.done = True
            return None
        return value

    def __dealloc__(self):
        delete_map_generator(self.generator)

cdef class VXLData:
    def __init__(self, fp = None):
        cdef unsigned char * c_data
        if fp is not None:
            data = fp.read()
            c_data = data
        else:
            c_data = NULL
        self.map = load_vxl(c_data)

    def load_vxl(self, c_data = None):
        self.map = load_vxl(c_data)

    def copy(self):
        cdef VXLData map = VXLData()
        map.map = copy_map(self.map)
        return map

    def get_point(self, int x, int y, int z):
        color = self.get_color(x, y, z)
        solid = color is not None
        return solid, color

    def set_point(self, int x, int y, int z, tuple color):
        if is_valid_position(x, y, z):
            set_point(x, y, z, self.map, 1, make_color(*color))

    # TODO: consider making this function raise error on invalid position
    cpdef get_solid(self, int x, int y, int z):
        if not is_valid_position(x, y, z):
            return None
        return get_solid(x, y, z, self.map)

    cpdef get_color(self, int x, int y, int z):
        if not self.get_solid(x, y, z):
            return None
        return make_color_tuple(get_color(x, y, z, self.map))

    def is_valid_position(self, int x, int y, int z):
        """return if the value is a valid position within the bounds of the
        map"""
        return is_valid_position(x, y, z)

    cpdef int get_z(self, int x, int y, int start = 0):
        '''
        Returns the first z coordinate that is solid beginning from `start` and
        moving down.  Useful for getting the coordinate for where something
        should be after being dropped.
        '''

        for z in range(start, 64):
            if get_solid(x, y, z, self.map):
                return z
        return 0

    cpdef int get_height(self, int x, int y):
        cdef int start = 63
        for z in range(start, -1, -1):
            if not get_solid(x, y, z, self.map):
                return z + 1
        return 0

    cpdef tuple get_safe_coords(self, int x, int y, int z):
        '''
        given (x, y, z) coords, return the closest set of coords on the map
        that is within the bounds of the map
        '''

        x = 0 if x < 0 else (511 if x > 511 else x)
        y = 0 if y < 0 else (511 if y > 511 else y)
        z = 0 if z < 0 else (63 if z > 63 else z)

        return (x, y, z)

    cpdef tuple get_random_point(self, int x1, int y1, int x2, int y2):
        cdef int x, y
        get_random_point(x1, y1, x2, y2, self.map, random.random(),
            random.random(), &x, &y)
        return x, y

    def count_land(self, int x1, y1, x2, y2):
        cdef int land = 0
        for x in range(x1, x2):
            for y in range(y1, y2):
                if self.get_solid(x, y, 62):
                    land += 1
        return land

    def destroy_point(self, int x, int y, int z):
        if not self.get_solid(x, y, z) or z >= 62:
            return 0
        set_point(x, y, z, self.map, 0, 0)
        count = 1
        start = time.monotonic()
        for node_x, node_y, node_z in self.get_neighbors(x, y, z):
            if node_z < 62:
                count += self.check_node(node_x, node_y, node_z, True)
        taken = time.monotonic() - start
        if taken > 0.1:
            print('destroying block at', x, y, z, 'took:', taken)
        return count

    def remove_point(self, int x, int y, int z):
        if is_valid_position(x, y, z):
            set_point(x, y, z, self.map, 0, 0)

    cpdef bint has_neighbors(self, int x, int y, int z):
        return (
            self.get_solid(x + 1, y, z) or
            self.get_solid(x - 1, y, z) or
            self.get_solid(x, y + 1, z) or
            self.get_solid(x, y - 1, z) or
            self.get_solid(x, y, z + 1) or
            self.get_solid(x, y, z - 1)
        )

    cpdef bint is_surface(self, int x, int y, int z):
        return (
            not self.get_solid(x, y, z - 1) or
            not self.get_solid(x, y, z + 1) or
            not self.get_solid(x + 1, y, z) or
            not self.get_solid(x - 1, y, z) or
            not self.get_solid(x, y + 1, z) or
            not self.get_solid(x, y - 1, z)
        )

    cpdef list get_neighbors(self, int x, int y, int z):
        cdef list neighbors = []
        for (node_x, node_y, node_z) in (
            (x, y, z - 1),
            (x, y - 1, z),
            (x, y + 1, z),
            (x - 1, y, z),
            (x + 1, y, z),
            (x, y, z + 1)):
            if self.get_solid(node_x, node_y, node_z):
                neighbors.append((node_x, node_y, node_z))
        return neighbors

    cpdef int check_node(self, int x, int y, int z, bint destroy = False):
        return check_node(x, y, z, self.map, destroy)

    cpdef bint build_point(self, int x, int y, int z, tuple color):
        if not is_valid_position(x, y, z):
            return False
        if not self.has_neighbors(x, y, z) or z >= 62:
            return False
        r, g, b = color
        set_point(x, y, z, self.map, 1, make_color(*color))
        return True

    cpdef bint set_column_fast(self, int x, int y, int z_start,
        int z_end, int z_color_end, int color):
        """Set a column's solidity, but only color a limited amount from
            the top."""
        if (not is_valid_position(x, y, z_start) or
            not is_valid_position(x, y, z_end) or
            z_end < z_start):
            return False
        set_column_solid(x, y, z_start, z_end, self.map, 1)

        if not is_valid_position(x, y, z_color_end) or z_color_end < z_start:
            return False
        set_column_color(x, y, z_start, z_color_end, self.map, color)
        return True

    cpdef update_shadows(self):
        update_shadows(self.map)

    def get_overview(self, int z = -1, bint rgba = False):
        cdef unsigned int * data
        cdef unsigned int i, r, g, b, a, color
        data_python = allocate_memory(sizeof(int[512][512]), <char**>&data)
        i = 0
        cdef int current_z
        if z == -1:
            a = 255
        else:
            current_z = z
        for y in range(512):
            for x in range(512):
                if z == -1:
                    current_z = self.get_z(x, y)
                else:
                    if get_solid(x, y, z, self.map):
                        a = 255
                    else:
                        a = 0
                color = get_color(x, y, current_z, self.map)
                if rgba:
                    b = color & 0xFF
                    g = (color & 0xFF00) >> 8
                    r = (color & 0xFF0000) >> 16
                    data[i] = r | (g << 8) | (b << 16) | (a << 24)
                else:
                    data[i] = (color & 0x00FFFFFF) | (a << 24)
                i += 1
        return data_python

    def set_overview(self, data_str, int z):
        cdef unsigned int * data
        cdef unsigned int r, g, b, a, color, i, new_color
        data = <unsigned int*>(<char*>data_str)
        i = 0
        for y in range(512):
            for x in range(512):
                color = data[i]
                a = (color & <unsigned int>0xFF000000) >> 24
                if a != 255:
                    set_point(x, y, z, self.map, 0, 0)
                else:
                    set_point(x, y, z, self.map, 1, color)
                i += 1

    def generate(self):
        start = time.monotonic()
        data = save_vxl(self.map)
        dt = time.monotonic() - start
        if dt > 1.0:
            print('VXLData.generate() took {}'.format(dt))
        return data

    def get_generator(self):
        return Generator(self)

    def __dealloc__(self):
        cdef MapData * map
        if self.map != NULL:
            map = self.map
            self.map = NULL
            delete_vxl(map)



"""
babel.py last modified 2014-04-04 22:12:44
Originally onectf.py by Yourself
Modified to complement babel_script.py: http://aloha.pk/files/aos/pyspades/feature_server/scripts/babel_script.py

Release thread:
http://www.buildandshoot.com/viewtopic.php?t=2586
"""

from pyspades.constants import CTF_MODE
from pyspades.collision import vector_collision

FLAG_SPAWN_POS = (256, 256)

HIDE_POS = (0, 0, 63)

DISABLED, ONE_CTF, REVERSE_ONE_CTF = range(3)

ONE_CTF_MODE = ONE_CTF

BABEL_CTF_MESSAGE = 'Take the intel to the enemy base to score.'


def apply_script(protocol, connection, config):

    class OneCTFConnection(connection):
        def on_flag_take(self):
            if self.protocol.one_ctf or self.protocol.reverse_one_ctf:
                flag = self.team.flag
                if flag.player is None:
                    flag.set(*HIDE_POS)
                    flag.update()
                    if self.protocol.reverse_one_ctf:
                        self.send_chat(BABEL_CTF_MESSAGE)
                else:
                    return False
            return connection.on_flag_take(self)

        def on_flag_drop(self):
            if self.protocol.one_ctf or self.protocol.reverse_one_ctf:
                flag = self.team.flag
                position = self.world_object.position
                x, y, z = int(
                    position.x), int(
                    position.y), max(
                    0, int(
                        position.z))
                z = self.protocol.map.get_z(x, y, z)
                flag.set(x, y, z)
                flag.update()
            return connection.on_flag_drop(self)

        def on_position_update(self):
            if self.protocol.reverse_one_ctf:
                if vector_collision(
                        self.world_object.position,
                        self.team.other.base):
                    other_flag = self.team.other.flag
                    if other_flag.player is self:
                        connection.capture_flag(self)
            return connection.on_position_update(self)

        def capture_flag(self):
            if self.protocol.reverse_one_ctf:
                self.send_chat(BABEL_CTF_MESSAGE)
                return False
            return connection.capture_flag(self)

        def on_flag_capture(self):
            if self.protocol.one_ctf or self.protocol.reverse_one_ctf:
                self.protocol.onectf_reset_flags()
            return connection.on_flag_capture(self)

    class OneCTFProtocol(protocol):
        game_mode = CTF_MODE
        one_ctf = False
        reverse_one_ctf = False

        def onectf_reset_flag(self, flag):
            z = self.map.get_z(*self.one_ctf_spawn_pos)
            pos = (self.one_ctf_spawn_pos[0], self.one_ctf_spawn_pos[1], z)
            if flag is not None:
                flag.player = None
                flag.set(*pos)
                flag.update()
            return pos

        def onectf_reset_flags(self):
            if self.one_ctf or self.reverse_one_ctf:
                self.onectf_reset_flag(self.blue_team.flag)
                self.onectf_reset_flag(self.green_team.flag)

        def on_game_end(self):
            if self.one_ctf or self.reverse_one_ctf:
                self.onectf_reset_flags()
            return protocol.on_game_end(self)

        def on_map_change(self, map_):
            self.one_ctf = self.reverse_one_ctf = False
            self.one_ctf_spawn_pos = FLAG_SPAWN_POS
            extensions = self.map_info.extensions
            if ONE_CTF_MODE == ONE_CTF:
                self.one_ctf = True
            elif ONE_CTF_MODE == REVERSE_ONE_CTF:
                self.reverse_one_ctf = True
            elif 'one_ctf' in extensions:
                self.one_ctf = extensions['one_ctf']
            if not self.one_ctf and 'reverse_one_ctf' in extensions:
                self.reverse_one_ctf = extensions['reverse_one_ctf']
            if 'one_ctf_spawn_pos' in extensions:
                self.one_ctf_spawn_pos = extensions['one_ctf_spawn_pos']
            return protocol.on_map_change(self, map_)

        def on_flag_spawn(self, x, y, z, flag, entity_id):
            pos = self.onectf_reset_flag(flag.team.other.flag)
            protocol.on_flag_spawn(
                self, pos[0], pos[1], pos[2], flag, entity_id)
            return pos

    return OneCTFProtocol, OneCTFConnection



"""
Makes grenades create blocks.

.. codeauthor:: hompy
"""

from pyspades.contained import BlockAction
from pyspades.constants import *


def apply_script(protocol, connection, config):
    def try_add_node(map, x, y, z, list):
        if x < 0 or x >= 512 or y < 0 or y >= 512 or z < 0 or z >= 62:
            return
        if map.get_solid(x, y, z):
            return
        list.append((x, y, z))

    class DirtGrenadeConnection(connection):

        def grenade_exploded(self, grenade):
            if self.name is None:
                return
            if self.weapon != 1:
                return connection.grenade_exploded(self, grenade)
            position = grenade.position
            x = int(position.x)
            y = int(position.y)
            z = int(position.z)
            blocks = 19
            map = self.protocol.map
            list = []
            try_add_node(map, x, y, z, list)
            block_action = BlockAction()
            block_action.value = BUILD_BLOCK
            block_action.player_id = self.player_id
            while list:
                x, y, z = list.pop(0)
                if connection.on_block_build_attempt(self, x, y, z) == False:
                    continue
                block_action.x = x
                block_action.y = y
                block_action.z = z
                self.protocol.broadcast_contained(block_action, save=True)
                map.set_point(x, y, z, self.color)
                blocks -= 1
                if blocks == 0:
                    break
                try_add_node(map, x, y, z - 1, list)
                try_add_node(map, x, y - 1, z, list)
                try_add_node(map, x, y + 1, z, list)
                try_add_node(map, x - 1, y, z, list)
                try_add_node(map, x + 1, y, z, list)
                try_add_node(map, x, y, z + 1, list)
            self.protocol.update_entities()

    return protocol, DirtGrenadeConnection



import abc

from pyspades import contained as loaders
from pyspades.collision import vector_collision, collision_3d
from pyspades.constants import TC_CAPTURE_DISTANCE


class BaseGamemode(metaclass=abc.ABCMeta):
    """Base Game Mode. You should probably not inherit directly from this,
    unless you enjoy doing low-level protocol fumbling."""

    @abc.abstractmethod
    def get_state_loader(self):
        """called to get the loader that contains the data to be sent after the
        state data, informing the client of the game mode and it's state"""


class IntelBasedGamemode(BaseGamemode):
    """A game mode where two team each have an intel briefcase and a tent.
    Often, the goal is to bring the enemy teams intel to the tent. This class
    only implements the required packet sending, not any actual game mode
    logic"""
    name = "ctf"

    def __init__(self, protocol):
        self.protocol = protocol
        self.team_1 = protocol.team_1
        self.team_2 = protocol.team_2

        self.state_loader = loaders.CTFState()
        self.intel_drop_loader = loaders.IntelDrop()
        self.intel_pickup_loader = loaders.IntelPickup()
        self.intel_capture_loader = loaders.IntelCapture()

        # player carrying team_1's intel, not the team_1 member carrying the
        # intel
        # This should probably be reversed as it's pretty confusing
        self.team_1_carrier = None
        self.team_2_carrier = None

    def on_position_update(self, player):
        target_flag = self.get_target_flag(player)
        if vector_collision(player.world_object.position,
                            player.team.base):
            if target_flag.player is self:
                player.capture_flag()
            player.check_refill()
        if target_flag.player is None and vector_collision(
                player.position, target_flag):
            player.take_flag()

    def get_state_loader(self):
        ctf_data = self.state_loader
        ctf_data.cap_limit = self.capture_limit
        ctf_data.team1_score = self.team_1.score
        ctf_data.team2_score = self.team_2.score

        team_1_base = self.team_1.base
        team_2_base = self.team_2.base

        ctf_data.team1_base_x = team_1_base.x
        ctf_data.team1_base_y = team_1_base.y
        ctf_data.team1_base_z = team_1_base.z

        ctf_data.team2_base_x = team_2_base.x
        ctf_data.team2_base_y = team_2_base.y
        ctf_data.team2_base_z = team_2_base.z

        team_1_intel = self.team_1.intel
        team_2_intel = self.team_2.intel

        if self.team_1_carrier is None:
            ctf_data.team1_has_intel = 0
            ctf_data.team2_flag_x = team_1_intel.x
            ctf_data.team2_flag_y = team_1_intel.y
            ctf_data.team2_flag_z = team_1_intel.z
        else:
            ctf_data.team1_has_intel = 1
            ctf_data.team2_carrier = self.team_2_carrier.player_id

        if self.team_2_carrier is None:
            ctf_data.team2_has_intel = 0
            ctf_data.team1_flag_x = team_2_intel.x
            ctf_data.team1_flag_y = team_2_intel.y
            ctf_data.team1_flag_z = team_2_intel.z
        else:
            ctf_data.team2_has_intel = 1
            ctf_data.team1_carrier = self.team_1_carrier.player_id

        return ctf_data

    @property
    @abc.abstractmethod
    def capture_limit(self):
        pass

    def on_player_reset(self, player):
        flag = self.get_player_flag(player)

        if flag is None:
            return

        position = player.position
        x = int(position.x)
        y = int(position.y)
        z = max(0, int(position.z))
        z = self.protocol.map.get_z(x, y, z)
        flag.set(x, y, z)
        flag.player = None
        intel_drop = loaders.IntelDrop()
        intel_drop.player_id = player.player_id
        intel_drop.x = flag.x
        intel_drop.y = flag.y
        intel_drop.z = flag.z
        self.protocol.broadcast_contained(intel_drop, save=True)
        player.on_flag_drop()

    def get_player_flag(self, player):
        for flag in (self.blue_flag, self.green_flag):
            if flag.player is self:
                return flag

        return None

    def get_target_flag(self, connection):
        return connection.team.other_flag


class TerritoryBasedGamemode(BaseGamemode):
    name = "tc"

    def __init__(self, protocol):
        self.protocol = protocol
        self.control_points = self.make_control_points()
        self.state_loader = loaders.TCState()

    @abc.abstractmethod
    def make_control_points():
        pass

    def get_state_packet(self):
        tc_data = self.state_loader
        tc_data.set_entities(self.control_points)
        return tc_data

    def on_position_update(self, connection):
        for entity in self.protocol.entities:
            collides = vector_collision(
                entity, connection.world_object.position, TC_CAPTURE_DISTANCE)
            if self in entity.players:
                if not collides:
                    entity.remove_player(self)
            else:
                if collides:
                    entity.add_player(self)
            if collides and vector_collision(entity,
                                             connection.world_object.position):
                connection.check_refill()



"""
BF-like squad system.

Commands
^^^^^^^^

* ``/squad <key>`` to join a squad
* ``/follow <player>`` to spawn near a specific player

Options
^^^^^^^

.. code-block:: guess

    [squad]
    respawn_time = "10sec"
    auto_squad = true

.. codeauthor:: Triplefox
"""

import random
from piqueserver.commands import command, get_player
from piqueserver import commands
from piqueserver.config import config, cast_duration
from piqueserver.server import respawn_time_option

SQUAD_NAMES = set([
    'Alpha', 'Bravo', 'Charlie', 'Delta', 'Epsilon', 'Foxtrot', 'Gamma',
    'Golf', 'Hotel', 'India', 'Juliet', 'Kilo', 'Lima', 'Mike',
    'November', 'Oscar', 'Papa', 'Quebec', 'Romero', 'Sierra', 'Tango',
    'Uniform', 'Victor', 'Whiskey', 'X-ray', 'Yankee', 'Zulu'])

squad_config = config.section('squad')
RESPAWN_TIME_OPTION = squad_config.option('respawn_time',
        default=respawn_time_option.get(), cast=cast_duration)
SIZE_OPTION = squad_config.option('size', 0)
AUTO_SQUAD_OPTION = squad_config.option('auto_squad', True)


@command()
def follow(self, playerkey=None):
    """
    Lets you spawn near a specific player in your squad
    /follow <player>
    """
    if playerkey is None:
        squad_pref = None
        squad = self.squad
    else:
        squad_pref = get_player(self.protocol, playerkey)
        squad = squad_pref.squad
        if squad_pref.team is not self.team:
            return '%s is not on your team!' % (squad_pref.name)
        if squad_pref is self:
            return "You can't follow yourself!"
        if squad_pref.squad is None:
            return ('%s is not in a squad and cannot be followed.' %
                    squad_pref.name)

    return self.join_squad(squad, squad_pref)


@command()
def squad(self, squadkey=None):
    """
    Lets you join a squad. If you die you respawn near your squad mates.
    /squad <key>
    """
    if self.protocol.squad_size <= 1:
        return 'Squads are disabled on this server.'

    # squadlist

    if squadkey is None:
        allsquads = self.get_squads(self.team)
        result = []
        for squadkey in list(allsquads.keys()):
            result.append(self.print_squad(
                squadkey, allsquads[squadkey]))
        result.append(('To join squads: /squad <squad name>. ' +
                       '/squad none to spawn normally.'))
        self.send_lines(result)
        return

    if squadkey.lower() == 'none':
        squad = None
        squad_pref = None
    else:
        squad = squadkey
        squad_pref = None

    return self.join_squad(squad, squad_pref)


def apply_script(protocol, connection, config):
    protocol.squad_respawn_time = RESPAWN_TIME_OPTION.get()
    protocol.squad_size = SIZE_OPTION.get()
    protocol.auto_squad = AUTO_SQUAD_OPTION.get()

    class SquadConnection(connection):
        squad = None
        squad_pref = None

        def on_login(self, name):
            if self.protocol.auto_squad:
                self.join_squad(self.find_auto_squad(), None)
            return connection.on_login(self, name)

        def get_squad(self, team, squadkey):
            result = {'name': squadkey, 'players': []}
            if squadkey is None:
                for player in list(self.protocol.players.values()):
                    if (player.team is team and
                            player.squad is None):
                        result['players'].append(player)
                        result['name'] = player.squad
            else:
                for player in list(self.protocol.players.values()):
                    if (player.team is team and player.squad and
                            player.squad.lower() == squadkey.lower()):
                        result['players'].append(player)
                        result['name'] = player.squad
            return result

        def get_squads(self, team):
            squad_dict = {}
            for player in list(self.protocol.players.values()):
                if player.team is team:
                    if player.squad in squad_dict:
                        squad_list = squad_dict[player.squad]
                    else:
                        squad_list = []
                        squad_dict[player.squad] = squad_list
                    squad_list.append(player)
            return squad_dict

        def print_squad(self, squadkey, squadlist):
            if squadkey is None:
                result = 'Unassigned: '
            else:
                result = 'Squad %s: ' % (squadkey)
            result += ', '.join([player.name for player in squadlist])
            return result

        def find_auto_squad(self):
            squad_dict = self.get_squads(self.team)
            available = set()
            unused = []
            for name in SQUAD_NAMES:
                if name in squad_dict:
                    if len(squad_dict[name]) < self.protocol.squad_size:
                        available.add(name)
                else:
                    unused.append(name)
            if available:
                return available.pop()
            else:
                return random.choice(unused)

        def join_squad(self, squad, squad_pref):

            if self.team is None or \
               self.team is self.protocol.spectator_team:
                return

            # same-squad check

            if squad is None or self.squad is None:
                newsquad = self.squad is not squad
            else:
                newsquad = self.squad.lower() != squad.lower()
            newpref = self.squad_pref is not squad_pref

            if not newsquad and not newpref:
                return 'Squad unchanged.'

            # unique squad, so check for squad size first

            existing = self.get_squad(self.team, squad)
            squad = existing['name']  # fixes the case

            if squad and (self.protocol.squad_size
                          <= len(existing['players'])):
                return ('Squad %s is full. (limit %s)' %
                        (squad, self.protocol.squad_size))

            # assign to unique squad

            oldsquad = self.squad
            oldpref = self.squad_pref

            if newsquad and self.squad:
                self.leave_squad()

            self.squad = squad
            self.squad_pref = squad_pref

            if newsquad and squad:
                self.squad_broadcast('%s joined your squad.' %
                                     self.name)

            if squad is None:
                self.respawn_time = self.protocol.respawn_time
                self.squad_pref = None
                self.send_chat('You are no longer assigned to a squad.')
            else:
                self.respawn_time = self.protocol.squad_respawn_time
                if newpref and newsquad:
                    if squad_pref is None:
                        return ('You are now in squad %s.' % squad)
                    else:
                        return ('You are now in squad %s, following %s.' %
                                (squad, squad_pref.name))
                elif newpref:
                    if squad_pref is None:
                        return ('You are no longer following %s.' %
                                oldpref.name)
                    else:
                        return ('You are now following %s.' %
                                squad_pref.name)
                elif newsquad:
                    return 'You are now in squad %s.' % squad

        def leave_squad(self):
            if self.squad:
                self.squad_broadcast("%s left your squad." %
                                     self.name)
            self.squad = None
            self.squad_pref = None
            for player in list(self.protocol.players.values()):
                if player.squad_pref is self:
                    player.squad_pref = None
            self.respawn_time = self.protocol.respawn_time

        def squad_broadcast(self, msg):
            if self.squad:
                squad = self.get_squad(self.team, self.squad)
                for player in squad['players']:
                    if player is not self:
                        player.send_chat(msg)

        def get_follow_location(self, follow):
            x, y, z = (follow.world_object.position.get())
            return x, y, z

        def on_team_changed(self, old_team):
            self.leave_squad()
            return connection.on_team_changed(self, old_team)

        def on_spawn(self, pos):
            if self.squad:
                all_members = ([n for n in self.get_squad(
                    self.team, self.squad)['players'] if n is not self])
                live_members = [
                    n for n in all_members if n.hp and not n.invisible and not n.god]
                membernames = [m.name for m in all_members]
                memberstr = ""
                for n in range(len(all_members)):
                    name = membernames[n]
                    if not all_members[n].hp:
                        name += " (DEAD)"
                    else:
                        name += " (%s hp)" % all_members[n].hp
                    if n == 0:
                        memberstr += "%s" % name
                    elif n == len(membernames) - 1:
                        memberstr += " and %s" % name
                    else:
                        memberstr += ", %s" % name
                if len(all_members) > 0:
                    self.send_chat('You are in squad %s with %s.' %
                                   (self.squad, memberstr))
                else:
                    self.send_chat('You are in squad %s, all alone.' %
                                   self.squad)
                if (self.squad_pref is not None and self.squad_pref.hp and
                    self.squad_pref.team is self.team and not self.squad_pref.invisible
                        and not self.squad_pref.god):
                    self.set_location_safe(self.get_follow_location(
                        self.squad_pref))
                else:
                    if len(live_members) > 0:
                        self.set_location_safe(self.get_follow_location(
                            random.choice(live_members)))
            return connection.on_spawn(self, pos)

        def on_kill(self, killer, type, grenade):
            if killer is None or killer is self:
                self.squad_broadcast('Squadmate %s suicided' % self.name)
            else:
                self.squad_broadcast('Squadmate %s was killed by %s' %
                                     (self.name, killer.name))
            return connection.on_kill(self, killer, type, grenade)

        def on_chat(self, value, global_message):
            if self.squad is not None and not global_message:
                value = '%s : %s' % (self.squad, value)
            return connection.on_chat(self, value, global_message)

    return protocol, SquadConnection



from datetime import datetime
from typing import Optional, Dict, Any
from twisted.logger import Logger
import aiohttp
from packaging import version
from piqueserver.version import __version__


log = Logger()


async def fetch_latest_release() -> Dict[str, Any]:
    endpoint = "https://api.github.com/repos/piqueserver/piqueserver/releases/latest"
    async with aiohttp.ClientSession() as session:
        async with session.get(endpoint) as response:
            return await response.json()


def format_release(release: Dict[str, Any]) -> str:
    latest_version = release["tag_name"]
    date = datetime.strptime(release["published_at"], "%Y-%m-%dT%H:%M:%SZ")
    formated = date.strftime("%b %d %Y")
    # git.io url points towards /latest release page
    return "New release available: {} ({}): https://git.io/fjIDk".format(latest_version, formated)


async def check_for_releases() -> Optional[Dict[str, Any]]:
    """Checks for new release and returns it if new release is found."""

    log.debug("checking latest version")
    try:
        release = await fetch_latest_release()
    except IOError as e:
        log.warn("could not fetch latest version: {err}", err=e)
        return None

    latest_version = release["tag_name"]
    if version.parse(latest_version) > version.parse(__version__):
        return release
    return None



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import sys
from os import path

try:
    from twisted.cred import portal, checkers
    from twisted.conch import manhole, manhole_ssh
    from twisted.conch.ssh import keys
except ImportError as e:
    print("ERROR: piqueserver was not installed with the [ssh] option")
    print("but SSH was enabled in the settings")
    print(e)
    sys.exit(1)

from piqueserver.config import config


def create_remote_factory(namespace, users):
    realm = manhole_ssh.TerminalRealm()

    def create_remote_protocol(_):
        return manhole.ColoredManhole(namespace)

    realm.chainedProtocolFactory.protocolFactory = create_remote_protocol
    p = portal.Portal(realm)

    users = {key: value.encode() for key, value in users.items()}

    p.registerChecker(
        checkers.InMemoryUsernamePasswordDatabaseDontUse(**users))
    f = manhole_ssh.ConchFactory(p)
    ssh_key_base_path = path.join(config.config_dir, "ssh-keys")
    ssh_pubkey_path = path.join(ssh_key_base_path,
                                "ssh_host_rsa_key.pub")
    ssh_privkey_path = path.join(ssh_key_base_path,
                                 "ssh_host_rsa_key")
    try:
        f.publicKeys[b"ssh-rsa"] = keys.Key.fromFile(ssh_pubkey_path)
        f.privateKeys[b"ssh-rsa"] = keys.Key.fromFile(ssh_privkey_path)
    except FileNotFoundError:
        print("ERROR: You don't have any keys in the host key location")
        print("Generate one with:")
        print("  mkdir {}".format(ssh_key_base_path))
        print("  ssh-keygen -f {} -t rsa".format(ssh_privkey_path))
        print("make sure to specify no password")
        sys.exit(1)
    return f


ssh_config = config.section("ssh")
class RemoteConsole:

    def __init__(self, server):
        users = ssh_config.option("users", {})
        port = ssh_config.option("port", 38827)
        factory = create_remote_factory(locals(), users.get())
        server.listenTCP(port.get(), factory)



"""
Saves current map on shutdown (and optionally loads it again on startup)

Options
^^^^^^^

.. code-block:: guess

    [savemap]
    load_saved_map = false

.. codeauthor:: mat^2
"""

import os
from twisted.internet import reactor, threads
from twisted.internet.defer import ensureDeferred
from twisted.logger  import Logger
from pyspades.vxl import VXLData
from piqueserver.config import config

log = Logger()

savemap_config = config.section('savemap')
LOAD_SAVED_MAP_OPTION = savemap_config.option('load_saved_map', False)


def get_name(map_info):
    return '%s/%s.saved.vxl' % (os.path.join(config.config_dir, 'maps'),
                                map_info.rot_info.name)


def load_map_from_path(path):
    with open(path, 'rb') as f:
        return VXLData(f)


def apply_script(protocol, connection, config):
    class MapSaveProtocol(protocol):

        def __init__(self, *arg, **kw):
            protocol.__init__(self, *arg, **kw)
            reactor.addSystemEventTrigger('before', 'shutdown', self.save_map)

        def make_map(self, name):
            if not LOAD_SAVED_MAP_OPTION.get():
                # just do the normal operation
                return protocol.make_map(self, name)

            async def do_load():
                map_info = await protocol.make_map(self, name)
                cached_path = get_name(map_info)

                if not os.path.isfile(cached_path):
                    # no cache file. Just load the regular map
                    return map_info

                log.info("Loading saved map for {name} from {cached_path}",
                    name=name, cached_path=cached_path)
                map_info.data = await threads.deferToThread(
                    load_map_from_path, cached_path)
                log.info("Saved map loaded")
                return map_info

            # replace the map_info deferred with our own one that also
            # loads and replaces the map data
            return ensureDeferred(do_load)

        def save_map(self):
            with open(get_name(self.map_info), 'wb') as f:
                f.write(self.map.generate())
            log.info("Map saved to {}".format(get_name(self.map_info)))

    return MapSaveProtocol, connection



"""
Shows K/D ratio

.. note::
   "ratio" must be AFTER "votekick" in the config script list

Commands
^^^^^^^^

* ``/ratio <player>`` shows K/D ratio

.. codeauthor:: TheGrandmaster & mat^2
"""

from piqueserver.commands import command, get_player
from pyspades.constants import *

# True if you want to include the headshot-death ratio in the ratio
# NOTE: this makes the message overflow into two lines
HEADSHOT_RATIO = False

# List other types of kills as well
EXTENDED_RATIO = False

# "ratio" must be AFTER "votekick" in the config.txt script list
RATIO_ON_VOTEKICK = True
IRC_ONLY = False


@command()
def ratio(connection, user=None):
    msg = "You have"
    if user is not None:
        connection = get_player(connection.protocol, user)
        msg = "%s has"
        if connection not in connection.protocol.players.values():
            raise KeyError()
        msg %= connection.name
    if connection not in connection.protocol.players.values():
        raise KeyError()

    kills = connection.ratio_kills
    deaths = float(max(1, connection.ratio_deaths))
    headshotkills = connection.ratio_headshotkills
    meleekills = connection.ratio_meleekills
    grenadekills = connection.ratio_grenadekills

    msg += " a kill-death ratio of %.2f" % (kills / deaths)
    if HEADSHOT_RATIO:
        msg += ", headshot-death ratio of %.2f" % (headshotkills / deaths)
    msg += " (%s kills, %s deaths" % (kills, connection.ratio_deaths)
    if EXTENDED_RATIO:
        msg += ", %s headshot, %s melee, %s grenade" % (
            headshotkills, meleekills, grenadekills)
    msg += ")."
    return msg


def apply_script(protocol, connection, config):
    class RatioConnection(connection):
        ratio_kills = 0
        ratio_headshotkills = 0
        ratio_meleekills = 0
        ratio_grenadekills = 0
        ratio_deaths = 0

        def on_kill(self, killer, type, grenade):
            if killer is not None and self.team is not killer.team:
                if self != killer:
                    killer.ratio_kills += 1
                    killer.ratio_headshotkills += type == HEADSHOT_KILL
                    killer.ratio_meleekills += type == MELEE_KILL
                    killer.ratio_grenadekills += type == GRENADE_KILL

            self.ratio_deaths += 1
            return connection.on_kill(self, killer, type, grenade)

    class RatioProtocol(protocol):

        def on_votekick_start(self, instigator, victim, reason):
            result = protocol.on_votekick_start(
                self, instigator, victim, reason)
            if result is None and RATIO_ON_VOTEKICK:
                message = ratio(instigator, victim.name)
                if IRC_ONLY:
                    self.irc_say('* ' + message)
                else:
                    self.send_chat(message, irc=True)
            return result

    return RatioProtocol, RatioConnection



"""
Ever wanted a disco in Ace of Spades?

Commands
^^^^^^^^

* ``/disco`` toggles disco

.. codeauthor:: mat^2
"""

from twisted.internet.task import LoopingCall
from twisted.internet.reactor import callLater
import random

from piqueserver.commands import command, admin

DISCO_ON_GAME_END = True
# Time is in seconds
DISCO_ON_GAME_END_DURATION = 10.0


@command('disco', admin_only=True)
def toggle_disco(connection):
    connection.protocol.toggle_disco(True)


DISCO_COLORS = set([
    (235, 64, 0),
    (128, 232, 121),
    (220, 223, 12),
    (43, 72, 228),
    (216, 94, 231),
    (255, 255, 255)
])


def apply_script(protocol, connection, config):
    class DiscoProtocol(protocol):
        current_colors = None
        disco = False
        old_fog_color = None

        def __init__(self, *arg, **kw):
            protocol.__init__(self, *arg, **kw)
            self.disco_loop = LoopingCall(self.update_color)

        def update_color(self):
            if not self.current_colors:
                self.current_colors = DISCO_COLORS.copy()
            color = self.current_colors.pop()
            self.set_fog_color(color)

        def on_game_end(self):
            if not self.disco and DISCO_ON_GAME_END:
                self.toggle_disco(False)
                callLater(DISCO_ON_GAME_END_DURATION, self.stop_disco)
            return protocol.on_game_end(self)

        def stop_disco(self):
            if self.disco:
                self.toggle_disco(False)

        def toggle_disco(self, message=False):
            self.disco = not self.disco
            if self.disco:
                self.old_fog_color = self.fog_color
                self.disco_loop.start(0.3)
                if message:
                    self.send_chat('DISCO PARTY MODE ENABLED!')
            else:
                self.disco_loop.stop()
                if self.old_fog_color is not None:
                    self.set_fog_color(self.old_fog_color)
                if message:
                    self.send_chat('The party has been stopped.')
    return DiscoProtocol, connection



# Free for all script written by Yourself

from random import randint

from pyspades.constants import CTF_MODE

# If ALWAYS_ENABLED is False, free for all can still be enabled in the map
# metadata by setting the key 'free_for_all' to True in the extensions
# dictionary
ALWAYS_ENABLED = True

# If WATER_SPANS is True, then players can spawn in water
WATER_SPAWNS = False

HIDE_POS = (0, 0, 63)


def apply_script(protocol, connection, config):
    class FreeForAllProtocol(protocol):
        game_mode = CTF_MODE
        free_for_all = False
        old_friendly_fire = None

        def on_map_change(self, map):
            extensions = self.map_info.extensions
            if ALWAYS_ENABLED:
                self.free_for_all = True
            else:
                if 'free_for_all' in extensions:
                    self.free_for_all = extensions['free_for_all']
                else:
                    self.free_for_all = False
            if self.free_for_all:
                self.old_friendly_fire = self.friendly_fire
                self.friendly_fire = True
            else:
                if self.old_friendly_fire is not None:
                    self.friendly_fire = self.old_friendly_fire
                    self.old_friendly_fire = None
            return protocol.on_map_change(self, map)

        def on_base_spawn(self, x, y, z, base, entity_id):
            if self.free_for_all:
                return HIDE_POS
            return protocol.on_base_spawn(self, x, y, z, base, entity_id)

        def on_flag_spawn(self, x, y, z, flag, entity_id):
            if self.free_for_all:
                return HIDE_POS
            return protocol.on_flag_spawn(self, x, y, z, flag, entity_id)

    class FreeForAllConnection(connection):
        score_hack = False

        def on_spawn_location(self, pos):
            if not self.score_hack and self.protocol.free_for_all:
                while True:
                    x = randint(0, 511)
                    y = randint(0, 511)
                    z = self.protocol.map.get_z(x, y)
                    if z != 63 or WATER_SPAWNS:
                        break
                # Magic numbers taken from server.py spawn function
                z -= 2.4
                x += 0.5
                y += 0.5
                return (x, y, z)
            return connection.on_spawn_location(self, pos)

        def on_refill(self):
            if self.protocol.free_for_all:
                return False
            return connection.on_refill(self)

        def on_flag_take(self):
            if self.protocol.free_for_all:
                return False
            return connection.on_flag_take(self)

        def on_kill(self, by, type, grenade):
            # Switch teams to add score hack
            if by is not None and by.team is self.team and self is not by:
                self.score_hack = True
                pos = self.world_object.position
                self.set_team(self.team.other)
                self.spawn((pos.x, pos.y, pos.z))
                self.score_hack = False
            return connection.on_kill(self, by, type, grenade)

    return FreeForAllProtocol, FreeForAllConnection



"""
Makes the flag returnable in Quake-like fashion.

.. codeauthor:: mat^2 & learn_more
"""

from pyspades.collision import vector_collision
from pyspades.constants import CTF_MODE


def apply_script(protocol, connection, config):

    class ReturnConnection(connection):

        def on_flag_take(self):
            self.team.other.flag.out = True
            return connection.on_flag_take(self)

        def on_flag_capture(self):
            self.team.other.flag.out = False
            self.team.other.flag.start = self.team.other.flag.get()
            return connection.on_flag_capture(self)

        def on_position_update(self):
            if self.protocol.game_mode == CTF_MODE:
                flag = self.team.flag
                if flag.player is None and flag.out:
                    if vector_collision(self.world_object.position, flag):
                        flag.out = False
                        flag.set(*flag.start)
                        flag.update()
                        self.protocol.send_chat('%s intel was returned by %s!' % (
                            self.team.name, self.name), global_message=None)
            return connection.on_position_update(self)

    class ReturnProtocol(protocol):

        def set_map(self, map):
            protocol.set_map(self, map)
            if self.game_mode == CTF_MODE:
                for team in self.teams.values():
                    if team.spectator:
                        continue
                    team.flag.out = False
                    team.flag.start = team.flag.get()

    return ReturnProtocol, ReturnConnection



"""
Changes the damage values depending on distance.

Options
^^^^^^^

.. code-block:: guess

   [rangedamange.rifle]
   pct_per_block = 0 # percentage per block?
   multiplier = 1

   [rangedamange.smg]
   pct_per_block = 0
   multiplier = 1

   [rangedamange.shotgun]
   pct_per_block = 0
   multiplier = 1

.. codeauthor:: ?
"""

from pyspades.constants import (SHOTGUN_WEAPON, SMG_WEAPON, RIFLE_WEAPON)
from pyspades.collision import distance_3d_vector
from math import sqrt
from piqueserver.config import config

range_damage_config = config.section("rangedamange")
rifle_config = range_damage_config.section("rifle")
smg_config = range_damage_config.section("smg")
shotgun_config = range_damage_config.section("shotgun")

rifle_pct_per_block = rifle_config.option("pct_per_block", 0)
rifle_multiplier = rifle_config.option("multiplier", 1)

shotgun_pct_per_block = shotgun_config.option("pct_per_block", 2.5)
shotgun_multiplier = shotgun_config.option("multiplier", 2)

smg_pct_per_block = smg_config.option("pct_per_block", 1.5)
smg_multiplier = smg_config.option("multiplier", 1.2)


def apply_script(protocol, connection, config):

    class RangeDamageConnection(connection):

        def __init__(self, *arg, **kw):
            connection.__init__(self, *arg, **kw)

        def on_hit(self, hit_amount, hit_player, type, grenade):
            result = connection.on_hit(self, hit_amount, hit_player, type,
                                       grenade)
            if result == False:
                return False
            if grenade:  # Don't reduce damage when using grenade e.g. airstrike
                return connection.on_hit(self, hit_amount, hit_player, type, grenade)
            if self.world_object and hit_player.world_object:
                if result is not None:
                    hit_amount = result
                dist = distance_3d_vector(
                    self.world_object.position, hit_player.world_object.position)
                if self.weapon == RIFLE_WEAPON:
                    pct = (100 * rifle_multiplier.get()
                           - rifle_pct_per_block.get() * dist)
                elif self.weapon == SMG_WEAPON:
                    pct = (100 * smg_multiplier.get()
                           - smg_pct_per_block.get() * dist)
                elif self.weapon == SHOTGUN_WEAPON:
                    pct = (100 * shotgun_multiplier.get()
                           - shotgun_pct_per_block.get() * dist)
                pct = max(0, pct) / 100.0
                hit_amount = int(hit_amount * pct)
                return hit_amount

    return protocol, RangeDamageConnection



from piqueserver.commands import (
    command, _commands, has_permission, get_player, get_command_help, player_only, target_player)


@command()
@player_only
def streak(connection):
    """
    Tell your current kill streak
    /streak
    """
    return ('Your current kill streak is %s. Best is %s kills' %
            (connection.streak, connection.best_streak))


@command()
@target_player
def ping(connection, player):
    """
    Tell your current ping (time for your actions to be received by the server)
    /ping
    """
    ping = player.latency
    if connection is player:
        return 'Your ping is {} ms. Lower ping is better!'.format(player.latency)
    return "{}'s ping is {} ms".format(player.name, player.latency)


@command()
@player_only
def rules(connection):
    """
    Show you the server rules
    /rules
    """
    lines = connection.protocol.rules
    if lines is None:
        return
    connection.send_lines(lines)


@command()
def commands(connection):
    """
    Print all available commands
    /commands
    """
    # Compact and simple output for admins
    if connection.admin:
        names = [command.command_name for command in _commands.values()
                 if has_permission(command, connection)]
        return 'Available commands: %s' % (', '.join(names))
    # More helpful output for regular players
    lines = []
    for cmd in _commands.values():
        if not has_permission(cmd, connection):
            continue
        desc, _, _ = get_command_help(cmd)
        lines.append("/{} {}".format(cmd.command_name, desc))
    connection.send_lines(lines)


@command("help")
def help_command(connection, command_name=None):
    """
    Gives description and usage info for a command
    /help <command_name>
    """
    # Querying usage for a specific command
    if command_name:
        if command_name not in _commands:
            return 'Unknown command'
        command_func = _commands[command_name]
        desc, usage, _ = get_command_help(command_func)
        return 'Description: {}\n Usage: {}'.format(desc, usage)
    # Output help if present in config
    if connection.protocol.help:
        return connection.send_lines(connection.protocol.help)



"""
Allows reloading config on the fly

Commands
^^^^^^^^

* ``/reloadconfig`` reloads the config (also updates rights) *admin only*

.. codeauthor:: Danke
"""

import json
import os.path

from piqueserver import commands
from piqueserver.commands import command
from piqueserver.config import config

# TODO: build an entire config reload core script


@command(admin_only=True)
def reloadconfig(connection):
    try:
        extension = os.path.splitext(config.config_file)[1][1:]
        with open(config.config_file) as fobj:
            config.update_from_file(fobj, format_=extension.upper())
    except Exception as e:
        print('Error reloading config:', e)
        return 'Error reloading config. Check log for details.'
    connection.protocol.reapply_config()
    return 'Config reloaded!'


def apply_script(protocol, connection, config):
    class PassreloadProtocol(protocol):
        def reapply_config(self):
            commands.update_rights(config.get('rights', {}))

    return PassreloadProtocol, connection



import asyncio
from asyncio import Future
from twisted.internet.defer import ensureDeferred, Deferred


def as_future(d: Deferred):
    return d.asFuture(asyncio.get_event_loop())


def as_deferred(f: Future):
    return Deferred.fromFuture(asyncio.ensure_future(f))



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from pyspades.bytes import ByteWriter

import enet


class BaseConnection:
    disconnected = False
    timeout_call = None

    def __init__(self, protocol, peer):
        self.protocol = protocol
        self.peer = peer

    def timed_out(self):
        self.disconnect()

    def disconnect(self, data=0):
        if self.disconnected:
            return
        self.disconnected = True
        self.peer.disconnect(data)
        self.protocol.remove_peer(self.peer)
        self.on_disconnect()

    def loader_received(self, loader):
        raise NotImplementedError('loader_received() not implemented')

    def send_contained(self, contained, sequence=False):
        if self.disconnected:
            return
        if sequence:
            flags = enet.PACKET_FLAG_UNSEQUENCED
        else:
            flags = enet.PACKET_FLAG_RELIABLE
        data = ByteWriter()
        contained.write(data)
        packet = enet.Packet(bytes(data), flags)
        self.peer.send(0, packet)

    # events

    def on_connect(self):
        pass

    def on_disconnect(self):
        pass

    # properties

    @property
    def latency(self):
        return self.peer.roundTripTime


class BaseProtocol:
    connection_class = BaseConnection
    max_connections = 33
    is_client = False

    def __init__(self, port=None, interface=b'*',
                 update_interval=1 / 60.0):
        if port is not None and interface is not None:
            address = enet.Address(interface, port)
        else:
            address = None
        try:
            self.host = enet.Host(address, self.max_connections, 1)
        except MemoryError:
            # pyenet raises memoryerror when the enet host could not be created
            raise IOError("Failed  to Create Enet Host. Is the Port in use?")

        self.host.compress_with_range_coder()
        self.update_loop = LoopingCall(self.update)
        self.update_loop.start(update_interval, False)
        self.connections = {}
        self.clients = {}

    def connect(self, connection_class, host, port, version, channel_count=1,
                timeout=5.0):
        host = host.encode()
        peer = self.host.connect(enet.Address(host, port), channel_count,
                                 version)
        connection = connection_class(self, peer)
        connection.timeout_call = reactor.callLater(timeout,
                                                    connection.timed_out)
        self.clients[peer] = connection
        return connection

    def on_connect(self, peer):
        connection = self.connection_class(self, peer)
        self.connections[peer] = connection
        connection.on_connect()

    def on_disconnect(self, peer):
        try:
            connection = self.connections.pop(peer)
            connection.disconnected = True
            connection.on_disconnect()
        except KeyError:
            return

    def data_received(self, peer, packet):
        connection = self.connections[peer]
        connection.loader_received(packet)

    def remove_peer(self, peer):
        if peer in self.connections:
            del self.connections[peer]
        elif peer in self.clients:
            del self.clients[peer]
            self.check_client()

    def check_client(self):
        if self.is_client and not self.clients:
            self.update_loop.stop()
            self.update_loop = None
            self.host = None  # important for GC

    def update(self):
        try:
            while 1:
                if self.host is None:
                    return
                try:
                    event = self.host.service(0)
                except IOError:
                    break
                if event is None:
                    break
                event_type = event.type
                if event_type == enet.EVENT_TYPE_NONE:
                    break
                peer = event.peer
                is_client = peer in self.clients
                if is_client:
                    connection = self.clients[peer]
                    if event_type == enet.EVENT_TYPE_CONNECT:
                        connection.on_connect()
                        connection.timeout_call.cancel()
                    elif event_type == enet.EVENT_TYPE_DISCONNECT:
                        connection.on_disconnect()
                        del self.clients[peer]
                        self.check_client()
                    elif event.type == enet.EVENT_TYPE_RECEIVE:
                        connection.loader_received(event.packet)
                else:
                    if event_type == enet.EVENT_TYPE_CONNECT:
                        self.on_connect(peer)
                    elif event_type == enet.EVENT_TYPE_DISCONNECT:
                        self.on_disconnect(peer)
                    elif event.type == enet.EVENT_TYPE_RECEIVE:
                        self.data_received(peer, event.packet)
        except:
            # make sure the LoopingCall doesn't catch this and stops
            import traceback
            traceback.print_exc()



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import os
import importlib
import math
import random
import time
from typing import List, Optional, Union

from twisted.logger import Logger

from pyspades.vxl import VXLData
from piqueserver.config import config

log = Logger()


class MapNotFound(Exception):

    def __init__(self, the_map):
        self.map = the_map
        Exception.__init__(self, 'map %s does not exist' % the_map)

    def __nonzero__(self):
        return False


def check_rotation(maps: List[Union[str, 'RotationInfo']], load_dir:
                   Optional[str]=None) -> List['RotationInfo']:
    """
    Checks if provided maps exist in maps dir. and
    returns an array of RotationInfo objects for those maps.
    Raises MapNotFound exception if maps are not found.
    """
    if load_dir is None:
        load_dir = os.path.join(config.config_dir, 'maps')
    infos = []
    for the_map in maps:
        if not isinstance(the_map, RotationInfo):
            the_map = RotationInfo(the_map)
        infos.append(the_map)
        if (not os.path.isfile(the_map.get_map_filename(load_dir))
                and not os.path.isfile(the_map.get_meta_filename(load_dir))):
            raise MapNotFound(the_map)
    return infos


class Map:
    # pylint: disable=too-many-instance-attributes

    def __init__(self, rot_info: 'RotationInfo', load_dir: str) -> None:
        self.load_information(rot_info, load_dir)

        # we want to count how long a map load or generate takes
        start_time = time.monotonic()
        if self.gen_script:
            seed = rot_info.get_seed()
            self.name = '{} #{}'.format(rot_info.name, seed)
            log.info("Generating map '{mapname}'...", mapname=self.name)
            random.seed(seed)
            self.data = self.gen_script(rot_info.name, seed)
        else:
            log.info("Loading map '%s'..." % self.name)
            self.load_vxl(rot_info)

        log.info('Map loaded successfully. (took {duration:.2f}s)',
                 duration=time.monotonic() - start_time)

    def load_information(self, rot_info: 'RotationInfo', load_dir: str) -> None:
        path = rot_info.get_meta_filename(load_dir)
        namespace = 'piqueserver_internal_map_' + rot_info.name
        try:
            loader = importlib.machinery.SourceFileLoader(namespace, path)
            spec = importlib.util.spec_from_loader(loader.name, loader)
            info = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(info)
        except FileNotFoundError:
            log.error("Map info file not found {}".format(path))
            info = None
        except Exception as e:
            log.error("Error while loading map info: {!r}".format(e))
            info = None

        self.info = info
        self.load_dir = load_dir
        self.rot_info = rot_info
        self.gen_script = getattr(info, 'gen_script', None)
        if self.gen_script:
            self.short_name = rot_info.name
            self.name = rot_info.full_name
        else:
            self.name = getattr(info, 'name', self.rot_info.name)
            self.short_name = self.name
        self.author = getattr(info, 'author', '(unknown)')
        self.version = getattr(info, 'version', '1.0')
        self.description = getattr(info, 'description', '')
        self.extensions = getattr(info, 'extensions', {})
        self.script = getattr(info, 'apply_script', None)
        self.time_limit = getattr(info, 'time_limit', None)
        self.cap_limit = getattr(info, 'cap_limit', None)
        self.get_spawn_location = getattr(info, 'get_spawn_location', None)
        self.get_entity_location = getattr(info, 'get_entity_location', None)
        self.on_map_change = getattr(info, 'on_map_change', None)
        self.on_map_leave = getattr(info, 'on_map_leave', None)
        self.on_block_destroy = getattr(info, 'on_block_destroy', None)
        self.is_indestructable = getattr(info, 'is_indestructable', None)

    def apply_script(self, protocol, connection, config):
        if self.script is not None:
            protocol, connection = self.script(protocol, connection, config)
        return protocol, connection

    def load_vxl(self, rot_info):
        try:
            fp = open(rot_info.get_map_filename(self.load_dir), 'rb')
        except OSError:
            raise MapNotFound(rot_info.name)
        self.data = VXLData(fp)
        fp.close()


class RotationInfo:
    seed = None

    def __init__(self, name: str = "pyspades") -> None:
        self.full_name = name

        splitted = name.split("#")
        if len(splitted) > 1:  # user specified a seed
            name = splitted[0].strip()
            self.seed = int(splitted[1])
        self.name = name

    def get_seed(self) -> int:
        if self.seed is not None:
            return self.seed
        random.seed()
        return random.randint(0, int(math.pow(2, 31)))

    def get_map_filename(self, load_dir: str) -> str:
        return os.path.join(load_dir, '%s.vxl' % self.name)

    def get_meta_filename(self, load_dir: str) -> str:
        return os.path.join(load_dir, '%s.txt' % self.name)

    def __str__(self):
        return self.full_name
